#!/usr/bin/env python3

"""
Command-line WebSocket client for use with Mod9 ASR WebSocket server.

The first positional argument is the URI of the WebSocket server to connect to.
The second positional argument may specify JSON-formatted request options.

Audio data may be read from stdin and sent to the specified WebSocket server.
Meanwhile, messages received from the WebSocket server are printed on stdout.

Example usage with a pre-recorded file:
```
curl -sL mod9.io/hi.wav | client.py ws://localhost:9980 > results.jsonl
```

Example usage with live streaming audio:
```
sox -dqV1 -twav -r16000 -c1 -b16 - | client.py wss://mod9.io '{"partial":true}'
```


"""

import argparse
import asyncio
import json
import os
import sys
import stat

import aiofiles
import websockets

# This is a fairly small default, set as a lowest common denominator for all use cases.  The ASR
# Engine receives data with a network buffer size of 128 bytes, but also utilizes an internal audio
# buffer size which is reflected by its "latency" request option. For example, at a default latency
# of 0.24s: 16kHZ 16-bit linear PCM audio would buffer as 0.24 * 16000 * 16/8 = 7680 bytes.
# At 0.10s:  8kHz  8-bit mu-law PCM audio would buffer as 0.10 * 8000 * 8/8 = 800 bytes.
# As general guidance: this parameter can be much larger for batch processing of pre-recorded audio,
# or if a real-time streaming application appears to be incurring significant networking overhead.
AUDIO_MESSAGE_SIZE = 128

# Required for use with aiofiles; hopefully this is fairly portable on *nix systems.
STDIN_FILENAME = '/dev/stdin'


async def from_stdin_to_engine(websocket, message_size=AUDIO_MESSAGE_SIZE):
    """
    Read chunked audio from stdin and send as messages to Engine WebSocket.
    Return when stdin is closed, after sending an empty message.

    Args:
        websocket (websockets.server.WebSocketServerProtocol):
            Engine WebSocket to send to (expects empty-message termination).
        message_size (int):
            Read audio chunks of this size; affects streaming latency.

    Returns:
        None
    """
    # https://stackoverflow.com/questions/13442574/how-do-i-determine-if-sys-stdin-is-redirected
    mode = os.fstat(sys.stdin.fileno()).st_mode

    if stat.S_ISFIFO(mode):
        # Async read from piped stdin: https://stackoverflow.com/q/64303607/281536
        # NOTE: this doesn't work with redirected stdin.
        reader = asyncio.StreamReader()
        protocol = asyncio.StreamReaderProtocol(reader)
        await asyncio.get_event_loop().connect_read_pipe(lambda: protocol, sys.stdin.buffer)
    elif stat.S_ISREG(mode):
        # Async read from redirected stdin.
        # NOTE: this also works for piped stdin, but the STDIN_FILENAME may not be very portable.
        reader = await aiofiles.open(STDIN_FILENAME, mode='rb')
    else:
        # Send an empty message to terminate audio data, (i.e. indicate that no audio is sent).
        # This will be an error for requests specifying format=wav, but not for format=raw.
        # It may be desirable someday for recognize requests to send empty audio with an audio-uri.
        await websocket.send(b'')
        return

    while True:
        audio_chunk = await reader.read(message_size)
        if audio_chunk:
            await websocket.send(audio_chunk)
        else:
            break

    # Unlike asyncio.StreamReader(), aiofiles provides a file-like object that should be closed.
    if isinstance(reader, aiofiles.threadpool.binary.AsyncBufferedReader):
        reader.close()

    # Send an empty message to terminate audio data.
    await websocket.send(b'')


async def from_engine_to_stdout(websocket):
    """
    Receive reply messages from Engine WebSocket, and print to stdout.
    Return when the WebSocket connection is closed by the Engine.

    Args:
        websocket (websockets.server.WebSocketServerProtocol):
            WebSocket to receive from (expecting single-line JSON replies).

    Returns:
        None
    """
    while True:
        try:
            reply_message = await websocket.recv()
        except websockets.ConnectionClosedOK:
            break

        reply_json = reply_message.decode()
        print(reply_json)


async def communicate_with_engine(uri, options_json, message_size):
    async with websockets.connect(uri) as websocket:
        # Send JSON-formatted options as initial message.
        options_message = options_json.encode('utf-8')
        await websocket.send(options_message)

        # The client's protocol differs depending on whether it's a recognize request sending audio.
        try:
            command = json.loads(options_json).get('command', 'recognize')
        except Exception:
            # Proceed with the request, but let the Engine report its own error message.
            command = 'recognize'

        if command == 'recognize':
            # Send audio to Engine in an async task.
            # NOTE: if the client sends audio for a command that does not expect it, or the Engine
            #       closes the connection early (e.g. on error), then it may take 10 seconds for the
            #       WebSocket connection to be closed by the Engine.
            # TODO: add logic to the client to await/parse the first Engine reply before proceeding?
            send_audio = asyncio.get_event_loop().create_task(
                from_stdin_to_engine(websocket, message_size),
            )

        # Receive replies from Engine in an async task.
        recv_replies = asyncio.get_event_loop().create_task(
            from_engine_to_stdout(websocket),
        )

        # Wait for the Engine to send all its replies, and then close the connection.
        await recv_replies

        # Cancel any audio remaining to be sent, since the Engine can't accept it anymore.
        if command == 'recognize':
            send_audio.cancel()


def main():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        'uri',
        metavar='URI',
        help='URI of the WebSocket server.',
        default='ws://localhost:9980',
    )
    parser.add_argument(
        'options_json',
        metavar='OPTIONS_JSON',
        nargs='?',
        help='JSON-formatted request options to pass to ASR Engine.',
        default='{}'
    )
    parser.add_argument(
        '--message-size',
        help='Size of audio sent in each WebSocket message, affecting streaming latency.',
        default=AUDIO_MESSAGE_SIZE,
        type=int,
    )
    args = parser.parse_args()

    try:
        asyncio.get_event_loop().run_until_complete(
            communicate_with_engine(
                args.uri,
                args.options_json,
                args.message_size,
            )
        )
    except KeyboardInterrupt:
        sys.exit(130)  # Don't show Python traceback.
    except Exception:
        print(f"ERROR: Unable to communicate with {args.uri} ... check if server is accessible?")
        sys.exit(1)


if __name__ == '__main__':
    main()



"""
Utilities including functions to talk to the Mod9 ASR Engine TCP Server.
"""

import datetime
import io
from itertools import _tee as TeeGeneratorType
import json
import logging
import os
import requests
import socket
import threading
import time
from types import GeneratorType
from urllib.parse import urlparse
from urllib.parse import quote

import boto3
from botocore.exceptions import ClientError as AWSClientError
import google.auth
import google.auth.transport.requests
import google.cloud.storage
import google.resumable_media.requests
from packaging import version

from mod9.reformat import config


class Mod9IncompatibleEngineVersionError(Exception):
    pass


class Mod9InsufficientEngineCapacityError(Exception):
    pass


class Mod9UnexpectedEngineResponseError(Exception):
    pass


class Mod9EngineFirstResponseNotProcessingError(Mod9UnexpectedEngineResponseError):
    pass


class Mod9EngineResponseNotCompletedError(Mod9UnexpectedEngineResponseError):
    pass


class Mod9EngineCouldNotGetModelNameError(Mod9UnexpectedEngineResponseError):
    pass


class Mod9UnexpectedEngineStateError(Exception):
    pass


class Mod9DisabledAudioURISchemeError(Exception):
    pass


class Mod9BadRequestError(Exception):
    pass


class Mod9LimitOverrunError(Exception):
    pass


class Mod9CouldNotAccessFileError(Exception):
    pass


class PropagatingThread(threading.Thread):
    """
    Thread class that propagates any exception that occurred during
    target function execution.

    See: https://stackoverflow.com/a/31614591
    """

    def run(self):
        self.exc = None
        try:
            self.ret = self._target(*self._args, **self._kwargs)
        except BaseException as e:
            self.exc = e

    def join(self):
        super().join()
        if self.exc:
            raise self.exc
        return self.ret


def get_local_path(local_path_with_prefix):
    """
    Get absolute path to file, assuming it begins with 'file://'.

    Args:
        local_path_with_prefix (str):
            Prefixed, possibly non-absolute path to file.

    Returns:
        str:
            Un-prefixed absolute path to file.
    """

    local_path = local_path_with_prefix.replace('file://', '', 1)
    return os.path.abspath(os.path.expanduser(local_path))


def parse_wav_encoding(audio_settings):
    """
    Determine if audio_settings has WAV file header.

    Args:
        audio_settings (Union[None, str, TeeGeneratorType[bytes]]):
            Audio content or location.

    Returns:
        Union[bool, str]:
            False or truthy value, which may be the encoding name.
        int:
            Sample rate if WAV header successfully parsed, else 0.
        int:
            Channel count if WAV header is successfully parsed, else 0.
    """

    if audio_settings is None:
        return False, 0, 0
    elif isinstance(audio_settings, TeeGeneratorType):
        audio_settings_header = next(audio_settings)
        return extract_encoding_rate_channel_from_wav_header(audio_settings_header)
    elif isinstance(audio_settings, str):
        uri_scheme = urlparse(audio_settings).scheme
        if uri_scheme == 'http' or uri_scheme == 'https':
            headers = {'Range': 'bytes=0-28'}
            with requests.get(audio_settings, headers=headers) as r:
                r.raise_for_status()
                return extract_encoding_rate_channel_from_wav_header(r.content)
        # TODO: try to read the WAV header for gs://, s3://, and file:// schemes.
        else:
            # Will fail in the pathological case of a non-WAV file stored with a ``.wav`` suffix,
            #  and when a WAV file is stored without a ``.wav`` suffix.
            return audio_settings.endswith('.wav'), 0, 0
    else:
        raise TypeError('Expected type None or TeeGeneratorType or str.')


def extract_encoding_rate_channel_from_wav_header(wav_header):
    """
    Attempt to extract encoding, sample rate from possible WAV header.

    Args:
        wav_header (bytes):
            Possible WAV file header.

    Returns:
        Union[bool, str]:
            False or truthy value, which may be the encoding name.
        int:
            Sample rate if WAV header successfully parsed, else 0.
        int:
            Channel count if WAV header is successfully parsed, else 0.
    """
    if wav_header[:4] == b'RIFF' and wav_header[8:12] == b'WAVE':
        # Get the encoding type, which is always truthy.
        encoding_bytes = wav_header[20:22]
        if encoding_bytes == b'\x01\x00':
            wav_encoding = 'pcm_s16le'
        elif encoding_bytes == b'\x03\x00':
            wav_encoding = 'pcm_f32le'
        elif encoding_bytes == b'\x06\x00':
            wav_encoding = 'a-law'
        elif encoding_bytes == b'\x07\x00':
            wav_encoding = 'mu-law'
        else:
            # TODO: handle WAVE_FORMAT_EXTENSIBLE with subformats.
            wav_encoding = True
        # Get number of channels.
        channels = int.from_bytes(wav_header[22:24], byteorder='little')
        # Get the sample rate.
        sample_rate = int.from_bytes(wav_header[24:28], byteorder='little')
        return wav_encoding, sample_rate, channels
    else:
        return False, 0, 0


def camel_case_to_snake_case(camel_case_string):
    """
    Convert a camelCase string to a snake_case string.

    Args:
        camel_case_string (str):
            A string using lowerCamelCaseConvention.

    Returns:
        str:
            A string using snake_case_convention.
    """

    # https://stackoverflow.com/a/44969381
    return ''.join(['_' + c.lower() if c.isupper() else c for c in camel_case_string]).lstrip('_')


def snake_case_to_camel_case(snake_case_string):
    """
    Convert a snake_case string to a camelCase string. Initial '_' will
    cause capitalization.

    Args:
        snake_case_string (str):
            A string using snake_case_convention.

    Returns:
        str:
            A string using lowerCamelCase or UpperCamelCase convention.
    """

    lower_case_words = snake_case_string.split(sep='_')
    camel_case_string = lower_case_words[0] \
        + ''.join(word.capitalize() for word in lower_case_words[1:])
    return camel_case_string


def recursively_convert_dict_keys_case(dict_in, convert_key):
    """
    Convert all the keys in a (potentially) recursive dict using
    function convert_key.

    Args:
        dict_in (dict[str, Union[dict, str]]):
            Dict of dicts and strs with str keys.
        convert_key (Callable[[str], str]):
            Function to convert str to str, to be applied to keys.

    Returns:
        dict[str, str]:
            Dict of dicts and strs with str keys; same structure as
            dict_in, with unchanged values, but with values transformed
            according to convert_key.
    """
    if not isinstance(dict_in, dict):
        return dict_in
    dict_out = {convert_key(key): value for key, value in dict_in.items()}
    for key, value in dict_out.items():
        if isinstance(value, dict):
            dict_out[key] = recursively_convert_dict_keys_case(dict_out[key], convert_key)
        if isinstance(value, list):
            dict_out[key] = [
                recursively_convert_dict_keys_case(item, convert_key) for item in dict_out[key]
            ]
    return dict_out


def get_bucket_key_from_path(bucketed_path_with_prefix, prefix):
    """
    Get bucket and key from path, assuming it begins with given prefix.

    Args:
        bucketed_path_with_prefix (str):
            Prefixed path including bucket and key.
        prefix (str):
            Prefix to look for in bucketed_path_with_prefix.

    Returns:
        tuple:
            bucket_name (str):
                Parsed name of bucket.
            key_name (str):
                Parsed name of key.
    """

    bucket_key = bucketed_path_with_prefix.replace(prefix, '', 1)
    bucket_name, key_name = bucket_key.split(sep='/', maxsplit=1)
    return bucket_name, key_name


def convert_gs_uri_to_http_url(uri):
    """
    Convert URI, assumed to begin with 'gs://', to URL for use with
    Google Resumable Media.

    Args:
        uri (str):
            URI to Google Cloud Storage beginning with ``gs://``.

    Returns:
        str:
            URL to Google Cloud Storage usable with Google Resumable
            Media.
    """

    bucket, key = get_bucket_key_from_path(uri, 'gs://')

    # https://cloud.google.com/storage/docs/request-endpoints#encoding
    key = quote(key, safe='')

    url = f"https://storage.googleapis.com/download/storage/v1/b/{bucket}" \
          f"/o/{key}?alt=media"
    return url


def generator_producer(generator, sock):
    """
    Send contents of generator to given socket.

    Args:
        generator (Iterable[bytes]):
            Data to send to socket.
        sock (socket.socket):
            Socket to send to.

    Returns:
        None
    """

    for chunk in generator:
        for i in range(0, len(chunk), config.MAX_CHUNK_SIZE):
            sock.sendall(chunk[i:i+config.MAX_CHUNK_SIZE])


def file_producer(uri, sock):
    """
    Send contents of file in chunks to given socket.

    Args:
        uri (str):
            Prefixed path to local file to be sent to socket.
        sock (socket.socket):
            Socket to send to.

    Returns:
        None
    """

    with open(get_local_path(uri), 'rb') as fin:
        for chunk in iter(lambda: fin.read(config.MAX_CHUNK_SIZE), b''):
            sock.sendall(chunk)


def http_producer(uri, sock):
    """
    Send contents of file hosted at URL in chunks to given socket.

    Args:
        uri (str):
            Prefixed public URL file to be sent to socket.
        sock (socket.socket):
            Socket to send to.

    Returns:
        None
    """

    with requests.get(uri, stream=True) as r:
        r.raise_for_status()
        for chunk in r.iter_content(chunk_size=config.MAX_CHUNK_SIZE):
            sock.sendall(chunk)


def google_cloud_producer(uri, sock):
    """
    Send contents of file hosted on GCS in chunks to given socket.

    Args:
        uri (str):
            Prefixed Google Cloud Storage file to to be sent to socket.
            Note you must have access to the file with currently loaded
            Google credentials.
        sock (socket.socket):
            Socket to send to.

    Returns:
        None
    """

    url = convert_gs_uri_to_http_url(uri)

    # https://googleapis.dev/python/google-resumable-media/latest/resumable_media/requests.html#google.resumable_media.requests.ChunkedDownload
    ro_scope = 'https://www.googleapis.com/auth/devstorage.read_only'
    credentials, _ = google.auth.default(scopes=(ro_scope,))
    transport = google.auth.transport.requests.AuthorizedSession(credentials)

    chunk_start = 0
    total_bytes = float('inf')
    while chunk_start < total_bytes:
        # ChunkedDownload appends bytes to stream.
        #  Use new stream each chunk to avoid holding entire file in memory.
        with io.BytesIO() as f:
            download = google.resumable_media.requests.ChunkedDownload(
                url,
                config.GS_CHUNK_SIZE,
                f,
                start=chunk_start,
            )
            download.consume_next_chunk(transport)
            chunk = f.getvalue()
            sock.sendall(chunk)

            chunk_start += config.GS_CHUNK_SIZE
            total_bytes = download.total_bytes


def aws_s3_producer(uri, sock):
    """
    Send contents of file hosted on AWS S3 in chunks to given socket.

    Args:
        uri (str):
            Prefixed AWS S3 file to to be sent to socket. Note you must
            have access to the file with currently loaded AWS
            credentials.
        sock (socket.socket):
            Socket to send to.

    Returns:
        None
    """

    bucket, key = get_bucket_key_from_path(uri, 's3://')
    # https://stackoverflow.com/a/40854612
    # https://botocore.amazonaws.com/v1/documentation/api/latest/reference/response.html#botocore.response.StreamingBody.iter_chunks
    s3c = boto3.client('s3')
    for chunk in s3c.get_object(Bucket=bucket, Key=key)['Body'].iter_chunks(
        chunk_size=config.MAX_CHUNK_SIZE
    ):
        sock.sendall(chunk)


def _make_eof_producer(producer):
    """
    Send a special EOF byte sequence to terminate the request.

    Args:
        producer (Callable[[str, socket.socket], None]):
            Request producer.

    Returns:
        Callable[[str, socket.socket], None]:
            Request producer that sends 'END-OF-FILE' at end of request.
    """

    def new_producer(audio_input, sock):
        producer(audio_input, sock)
        sock.sendall(b'END-OF-FILE')
    return new_producer


def connect_to_engine(host=None, port=None):
    """
    Create a socket connection to Engine at ``host``, ``port``.

    Args:
        host (Union[str, None]):
            Engine host, or None to default to config.
        port (Union[int, None]):
            Engine port, or None to default to config.

    Returns:
        socket.socket:
            The socket connection to the Engine.
    """

    if host is None:
        host = config.ASR_ENGINE_HOST
    if port is None:
        port = config.ASR_ENGINE_PORT

    return socket.create_connection(
        (host, port),
        timeout=config.SOCKET_CONNECTION_TIMEOUT_SECONDS,
    )


prefix_producer_map = {
    'file':  file_producer,
    'gs':    google_cloud_producer,
    'http':  http_producer,
    'https': http_producer,
    's3':    aws_s3_producer,
}


def get_transcripts_mod9(
    options,
    audio_input,
    host=None,
    port=None,
):
    """
    Open TCP connection to Mod9 server, send input, and yield output
    generator.

    Args:
        options (dict[str, Union[dict, float, int, str]]):
            Transcription options for Engine.
        audio_input (Union[GeneratorType, TeeGeneratorType, str]):
            Audio content to be transcribed.
        host (Union[str, None]):
            Engine host, or None to default to config.
        port (Union[int, None]):
            Engine port, or None to default to config.

    Yields:
        dict[str, Union[dict, float, int, str]]:
            Result from Mod9 ASR Engine TCP Server.
    """

    with connect_to_engine(host=host, port=port) as sock:
        sock.settimeout(config.SOCKET_INACTIVITY_TIMEOUT_SECONDS)

        # Start by sending the options as JSON on the first line (terminated w/ newline character).
        first_request_line = json.dumps(options, separators=(',', ':')) + '\n'
        sock.sendall(first_request_line.encode())

        # The Engine should respond with an initial 'processing' status message.
        sockfile = sock.makefile(mode='r', encoding='utf-8')
        first_response_line = json.loads(sockfile.readline())
        if first_response_line.get('status') != 'processing':
            raise Mod9EngineFirstResponseNotProcessingError(
                f"Did not receive 'processing' from Mod9 ASR Engine. Got '{first_response_line}'."
            )

        # Select proper producer given audio input type.
        if isinstance(audio_input, GeneratorType) or isinstance(audio_input, TeeGeneratorType):
            producer = generator_producer
        elif isinstance(audio_input, str):
            producer = prefix_producer_map.get(urlparse(audio_input).scheme)
            if producer is None:
                allowed_prefixes = '://, '.join(prefix_producer_map) + '://'
                raise NotImplementedError(
                    f"URI '{audio_input}' has unrecognized prefix."
                    f" Allowed prefixes: {allowed_prefixes}."
                )
        else:
            raise TypeError(f"Audio input should be generator or str; got '{type(audio_input)}'.")

        if options.get('format') == 'raw':
            producer = _make_eof_producer(producer)

        # Launch producer thread to stream from audio input source to Engine.
        producer_thread = PropagatingThread(target=producer, args=(audio_input, sock))
        producer_thread.start()

        yield first_response_line
        for line in sockfile:
            yield json.loads(line)

        producer_thread.join()


def get_loaded_models_mod9(host=None, port=None):
    """
    Query Engine for a list of models and return loaded models.

    Args:
        host (Union[str, None]):
            Engine host, or None to default to config.
        port (Union[int, None]):
            Engine port, or None to default to config.

    Returns:
        dict[str, dict[str, Union[bool, dict, int, str]]]:
            Metadata about models currently loaded in Engine.
    """

    with connect_to_engine(host=host, port=port) as sock:
        sock.settimeout(config.SOCKET_INACTIVITY_TIMEOUT_SECONDS)

        # Start by sending request to list models (terminated w/ newline character).
        get_model_request = '{"command": "get-models-info"}\n'
        sock.sendall(get_model_request.encode())

        sockfile = sock.makefile(mode='r', encoding='utf-8')

        response_raw = sockfile.read()  # expected to be single-line JSON
        get_models_response = json.loads(response_raw)
        if get_models_response.get('status') != 'completed':
            raise Mod9EngineResponseNotCompletedError(
                f"Got response '{get_models_response}'; must have `.status` field `completed`."
            )
        if 'asr_models' not in get_models_response:
            raise Mod9UnexpectedEngineResponseError(
                f"Got response '{get_models_response}'; must have `.asr_models` field."
            )

        sockfile.close()

        return get_models_response['asr_models']


def find_loaded_models_with_rate(rate, host=None, port=None):
    """
    Get all models loaded in Engine that have the specified rate.

    Args:
        rate (int):
            Check for loaded Engine models with this rate.
        host (Union[str, None]):
            Engine host, or None to default to config.
        port (Union[int, None]):
            Engine port, or None to default to config.

    Returns:
        list[dict[str, Union[bool, dict, int, str]]]:
            Metadata about models currently loaded in Engine with specified rate (or empty list).
    """
    try:
        models = get_loaded_models_mod9(host=host, port=port)
    except Mod9UnexpectedEngineResponseError:
        return []
    return [model for model in models if model['rate'] == rate]


def find_loaded_models_with_language(language_code, host=None, port=None):
    """
    Get all models loaded in Engine with the specified language code prefix.

    Args:
        language_code (str):
            Check for loaded Engine models with the prefix of this language code.
        host (Union[str, None]):
            Engine host, or None to default to config.
        port (Union[int, None]):
            Engine port, or None to default to config.

    Returns:
        list[dict[str, Union[bool, dict, int, str]]]:
            Metadata about models currently loaded in Engine with
            specified language code prefix (or empty list).
    """

    try:
        models = get_loaded_models_mod9(host=host, port=port)
    except Mod9UnexpectedEngineResponseError:
        return []
    language_code_region = language_code.lower().split(sep='-')
    possible_models = []
    possible_model_language_code_regions = []
    for model in models:
        model_language_code = model.get('language')
        if model_language_code:
            model_language_code_region = model_language_code.lower().split(sep='-')
            if language_code_region[0] == model_language_code_region[0]:
                possible_models.append(model)
                possible_model_language_code_regions.append(model_language_code_region)
    return possible_models


def get_model_languages(host=None, port=None):
    """
    Get the language code of all models loaded in the Engine.

    Args:
        host (Union[str, None]):
            Engine host, or None to default to config.
        port (Union[int, None]):
            Engine port, or None to default to config.

    Returns:
        set[str]:
            Language code of all models loaded in the Engine.
    """

    try:
        models = get_loaded_models_mod9(host=host, port=port)
    except Mod9UnexpectedEngineResponseError:
        return set()
    languages = set()
    for model in models:
        model_language_code = model.get('language')
        if model_language_code:
            languages.add(model_language_code)
    return languages


def select_best_model_for_language_code(models, language_code, model_type=None):
    """
    Get model with language code that best matches given tag.

    Args:
        models (list[dict[str, Union[bool, dict, int, str]]]):
            Model metadata from which to select best match.
        language_code (str):
            Check for loaded Engine models with the prefix of this language code.

    Returns:
        dict[str, Union[bool, dict, int, str]]:
            Model corresponding to the best match (or first model).
    """

    def use_model_type(models_subset):
        if model_type:
            model_type_mod9 = 'phone' if model_type == 'phone_call' else 'video'
            for model in models_subset:
                if model_type_mod9 in model['name']:
                    return model
        return None

    language_code_region = language_code.lower().split(sep='-')
    model_language_code_regions = [model['language'].lower().split(sep='-') for model in models]
    possible_models = []
    if len(language_code_region) == 1:
        for i, model_language_code_region in enumerate(model_language_code_regions):
            if len(model_language_code_region) == 1:
                possible_models.append(models[i])
        if len(possible_models) == 1:
            return possible_models[0]
        elif len(possible_models) > 1:
            model_via_type = use_model_type(possible_models)
            if model_via_type:
                return model_via_type
        model_via_type = use_model_type(models)
        if model_via_type:
            return model_via_type
        return models[0]

    general_models = []
    model_type_models = []
    for i, model_language_code_region in enumerate(model_language_code_regions):
        if len(model_language_code_region) == 1:
            general_models.append(models[i])
        elif language_code_region[1] == model_language_code_region[1]:
            model_type_models.append(models[i])
    if len(model_type_models) == 1:
        return model_type_models[0]
    elif len(model_type_models) > 1:
        model_via_type = use_model_type(model_type_models)
        if model_via_type:
            return model_via_type
    if len(general_models) == 1:
        return general_models[0]
    elif len(general_models) > 1:
        model_via_type = use_model_type(general_models)
        if model_via_type:
            return model_via_type
    return models[0]


def get_version_mod9(host=None, port=None):
    """
    Query Engine for version number.

    Args:
        host (Union[str, None]):
            Engine host, or None to default to config.
        port (Union[int, None]):
            Engine port, or None to default to config.

    Returns:
        string:
            The version of the Engine.

    Raises:
        OSError:
            Socket errors.
        TypeError:
            Version parsing errors.
    """

    with connect_to_engine(host=host, port=port) as sock:
        sock.settimeout(config.SOCKET_INACTIVITY_TIMEOUT_SECONDS)

        sock.sendall('{"command": "get-version"}\n'.encode())

        with sock.makefile(mode='r', encoding='utf-8') as sockfile:
            response_raw = sockfile.read()  # expected to be single-line JSON
            response = json.loads(response_raw)

        return response.get('version')


def is_compatible_mod9(engine_version_string):
    """
    Determine if present wrappers are compatible with Engine version.

    Args:
        engine_version_string (Union[str, None]):
            The Engine version to compare to wrapper allowed range.

    Returns:
        bool:
            Whether the wrappers and Engine are compatible.

    Raises:
        OSError:
            Socket errors.
        ValueError:
            Invalid semantic version given to comparator.
    """

    engine_version = version.parse(engine_version_string)

    lower_bound_string, upper_bound_string = config.WRAPPER_ENGINE_COMPATIBILITY_RANGE

    is_within_lower_bound = True
    is_within_upper_bound = True

    if lower_bound_string is not None:
        lower_bound = version.parse(lower_bound_string)
        is_within_lower_bound = lower_bound <= engine_version  # Lower bound is inclusive.
    if upper_bound_string is not None:
        upper_bound = version.parse(upper_bound_string)
        is_within_upper_bound = engine_version < upper_bound  # Upper bound is exclusive.

    return is_within_lower_bound and is_within_upper_bound


def test_host_port(logger=None, host=None, port=None):
    """
    Check if Mod9 ASR Engine is online. Loop until get-info command
    provides a ``ready`` response. Log stats.

    Args:
        host (Union[str, None]):
            Engine host, or None to default to config.
        port (Union[int, None]):
            Engine port, or None to default to config.

    Returns:
        None
    """

    if not logger:
        logger = logging.root

    engine_version = get_version_mod9(host=host, port=port)
    if not is_compatible_mod9(engine_version):
        raise Mod9IncompatibleEngineVersionError(
            f"This Python SDK version {config.WRAPPER_VERSION} is not compatible with"
            f" ASR Engine version {engine_version} (which should be"
            f" >={config.WRAPPER_ENGINE_COMPATIBILITY_RANGE[0]}"
            + (').' if config.WRAPPER_ENGINE_COMPATIBILITY_RANGE[1] is None
               else f" and <{config.WRAPPER_ENGINE_COMPATIBILITY_RANGE[1]}).")
        )

    logger.info(
        "Checking for ASR Engine running at %s on port %d ...",
        config.ASR_ENGINE_HOST, config.ASR_ENGINE_PORT
    )

    # Loop sending get-info until receive ``state`` in response as "ready".
    response = dict()
    response_raw = ''
    while response.get('state') != 'ready':
        with connect_to_engine(host=host, port=port) as sock:
            sock.settimeout(config.SOCKET_INACTIVITY_TIMEOUT_SECONDS)

            sock.sendall('{"command": "get-info"}\n'.encode())
            with sock.makefile(mode='r', encoding='utf-8') as sockfile:
                response_raw = sockfile.read()  # expected to be single-line JSON
                response = json.loads(response_raw)

        # Log and sleep except when receiving a ready response.
        if response.get('state') == 'starting':
            logger.warning(
                "The Engine is still starting. Will attempt to connect again in %s seconds...",
                config.ENGINE_CONNECTION_RETRY_SECONDS,
            )
            time.sleep(config.ENGINE_CONNECTION_RETRY_SECONDS)
        elif response.get('state') != 'ready':
            raise Mod9UnexpectedEngineStateError(
                'Engine responded with unexpected state: ' + response.get('state')
            )

    if response['requests']['limit'] - response['requests']['active'] == 0:
        logger.warning('The Engine is at its limit, and unable to accept new requests.')

    logger.info("The ASR Engine is ready: %s", response_raw.strip())
    # TODO: also report the name of loaded models, from the get-models-info command?


def validate_uri_scheme(uri, allowed_uri_schemes):
    """
    Check if a URI has scheme in allowed set.

    Args:
        uri (str):
            URI with a scheme that may or may not be allowed.
        allowed_uri_schemes (set[str]):
            Set of schemes the URI is allowed to have.

    Raises:
        Mod9DisabledAudioURISchemeError:
            The error message indicating which schemes are allowed,
            if given URI had a different scheme, or None if it did not.
    """

    uri_scheme = urlparse(uri).scheme
    if uri_scheme not in allowed_uri_schemes:
        error_message = f"URI '{uri_scheme}' not allowed; server configured to allow:" + \
                        f" {', '.join(allowed_uri_schemes) if len(allowed_uri_schemes) else None}."

        raise Mod9DisabledAudioURISchemeError(error_message)


def format_log_time(self, record, datefmt):
    """
    Helper function to format logging record times in ISO-8601 format.

    Usage is with a ``logging.Formatter``, e.g.,
    ```
    logging.Formatter.formatTime = format_log_time
    ```

    See:
    https://stackoverflow.com/a/58777937

    Args:
        self (logging.Formatter):
            The log formatter calling this function.
        record (logging.LogRecord):
            The log record whose time to format.
        datefmt (Union[str, None]):
            Optional string to specify time format; overriden here.

    Returns:
        str:
            The time, formatted as 2006-01-02T15:04:05.999-07:00.
    """

    log_time = datetime.datetime.fromtimestamp(record.created, datetime.timezone.utc)
    return log_time.astimezone().isoformat(timespec='milliseconds')


def uri_exists(uri):
    """
    Attempt to access remote file at given uri.

    Args:
        uri (str):
            URI where remote file resides.

    Returns:
        bool:
            Whether the remote file is reachable.
    """

    parsed_uri = urlparse(uri)
    try:
        if parsed_uri.scheme == 'file':
            file_path = parsed_uri.netloc + parsed_uri.path
            if not os.path.exists(file_path):
                raise Mod9CouldNotAccessFileError
        elif parsed_uri.scheme == 'gs':
            bucket, key = get_bucket_key_from_path(uri, 'gs://')
            ro_scope = 'https://www.googleapis.com/auth/devstorage.read_only'
            credentials, _ = google.auth.default(scopes=(ro_scope,))
            gsc = google.cloud.storage.Client(credentials=credentials)
            blob = google.cloud.storage.Blob(bucket=gsc.bucket(bucket), name=key)
            if not blob.exists(client=gsc):
                raise Mod9CouldNotAccessFileError
        elif parsed_uri.scheme == 'http' or parsed_uri.scheme == 'https':
            with requests.head(uri) as r:
                r.raise_for_status()
        elif parsed_uri.scheme == 's3':
            bucket, key = get_bucket_key_from_path(uri, 's3://')
            s3c = boto3.client('s3')
            s3c.head_object(Bucket=bucket, Key=key)
    except (AWSClientError, requests.HTTPError, Mod9CouldNotAccessFileError) as e:
        raise Mod9CouldNotAccessFileError(f"Could not access file at {uri}.") from e



"""
Common functions used by the Google-compatible Mod9 ASR Python SDK
wrappers.
"""

import google.cloud.speech_v1p1beta1 as cloud_speech

from mod9.reformat.utils import get_transcripts_mod9
from mod9.reformat import google as reformat


class Mod9NotImplementedError(Exception):
    pass


def long_running_recognize(
    request,
    retry=None,
    timeout=None,
    metadata=None,
    module=cloud_speech,
    host=None,
    port=None,
    *args,
    **kwargs,
):
    """
    Performs asynchronous speech recognition.

    Mod9: Not currently implemented.
    """

    raise Mod9NotImplementedError('Refer to the REST wrapper for long-running recognition.')


def recognize(
    request,
    retry=None,
    timeout=None,
    metadata=None,
    module=cloud_speech,
    host=None,
    port=None,
    *args,
    **kwargs,
):
    """
    Performs synchronous speech recognition: receive results after all
    audio has been sent and processed.

    Args:
        request (Union[dict, RecognizeRequest]):
            The request object. The top-level message sent by the client
            for the `Recognize` method.
        config (RecognitionConfig):
            Required. Provides information to the recognizer that
            specifies how to process the request. This corresponds to
            the ``config`` field on the ``request`` instance; if
            ``request`` is provided, this should not be set.
        audio (RecognitionAudio):
            Required. The audio data to be recognized. This corresponds
            to the ``audio`` field on the ``request`` instance; if
            ``request`` is provided, this should not be set.
        module (module):
            Module to read Google-like types from, in case of
            subclassing. Default is ``google.cloud.speech_v1p1beta1``.
        host (Union[str, None]):
            Engine host, or None to default to config.
        port (Union[int, None]):
            Engine port, or None to default to config.

    Returns:
        RecognizeResponse:
            The only message returned to the client by the ``Recognize``
            method. It contains the result as zero or more sequential
            ``SpeechRecognitionResult`` messages.
    """

    if retry or timeout or metadata:
        # NOTE: Mod9 ASR Python SDK wrapper does not support retry,
        #  timeout, or metadata arguments, which are currently ignored.
        # TODO: Add support for retry and timeout arguments.
        pass

    # Parse inputs to ensure they are the expected encoding, have allowed arguments.
    options, requests = reformat.input_to_mod9(
        {'config': request.config, 'audio': request.audio},
        module=module,
        host=host,
        port=port,
    )

    # Read Engine responses.
    mod9_results = get_transcripts_mod9(
        options,
        requests,
        host=host,
        port=port,
    )

    # Convert Mod9 style to Google style.
    google_result_dicts = reformat.result_from_mod9(mod9_results, host=host, port=port)

    # Convert Mod9 type to Google type.
    google_results = reformat.google_type_result_from_dict(
        google_result_dicts,
        google_result_type=module.SpeechRecognitionResult,
        module=module,
    )

    # Yield the expected response object.
    response = module.RecognizeResponse()
    for google_result in google_results:
        response.results.append(google_result)

    return response


def streaming_recognize(
    requests,
    retry=None,
    timeout=None,
    metadata=None,
    module=cloud_speech,
    host=None,
    port=None,
    *args,
    **kwargs,
):
    """
    Perform streaming speech recognition.

    This method allows you to receive results while sending audio.

    Example:
      >>> from mod9.asr import speech
      >>> client = speech.SpeechClient()
      >>> encoding = speech.RecognitionConfig.AudioEncoding.LINEAR16
      >>> config = speech.StreamingRecognitionConfig(
      ...     config=speech.RecognitionConfig(
      ...         encoding=encoding,
      ...     ),
      ... )
      >>> # Byte-encoded content goes in ``audio_content`` below:
      >>> request = speech.StreamingRecognizeRequest(audio_content=b'')
      >>> requests = [request]
      >>> for element in client.streaming_recognize(config, requests):
      ...     # process element
      ...     pass

    Args:
        config (StreamingRecognitionConfig):
            The configuration to use for the stream.
        requests (Iterable[StreamingRecognizeRequest]):
            The input objects.
        module (module):
            Module to read Google-like types from, in case of
            subclassing. Default is ``google.cloud.speech_v1p1beta1``.
        host (Union[str, None]):
            Engine host, or None to default to config.
        port (Union[int, None]):
            Engine port, or None to default to config.

    Returns:
        Iterable[StreamingRecognizeResponse]
    """

    if retry or timeout or metadata:
        # NOTE: Mod9 ASR Python SDK wrapper does not support retry,
        #  timeout, or metadata arguments, which are currently ignored.
        # TODO: Add support for retry and timeout arguments.
        pass

    # The first request should have the config and no audio.
    request = next(requests)

    # Parse inputs to ensure they are the expected encoding, have allowed arguments.
    options, _ = reformat.input_to_mod9(
        {'config': request['streaming_config']},
        module=module,
        host=host,
        port=port,
    )

    # Read Engine responses.
    audio_requests = (request.audio_content for request in requests)
    mod9_results = get_transcripts_mod9(
        options,
        audio_requests,
        host=host,
        port=port,
    )

    # Convert Mod9 style to Google style.
    google_result_dicts = reformat.result_from_mod9(mod9_results, host=host, port=port)

    # Convert Mod9 type to Google type.
    google_results = reformat.google_type_result_from_dict(
        google_result_dicts,
        google_result_type=module.StreamingRecognitionResult,
        module=module,
    )

    # Yield the expected response object.
    for google_result in google_results:
        response = module.StreamingRecognizeResponse()
        response.results.append(google_result)
        yield response



#!/usr/bin/env python3

import argparse
import json
import logging
import sys
import uuid

from elasticsearch import Elasticsearch
from elasticsearch import RequestsHttpConnection
from elasticsearch.exceptions import RequestError

from mod9.reformat import utils

ES_HOSTS = ['localhost']
ES_PORT = 9200
ES_INDEX_CREATION_TIMEOUT_SECONDS = 120

TEMPLATE = {
    'mappings': {
        'date_detection': False,
        'properties': {
            'title': {
                'type': 'text',
            },
            'tracks': {
                'type': 'nested',
                'properties': {
                    'asr': {
                        'type': 'text',
                        'analyzer': 'analyzer-cats-index',
                        'search_analyzer': 'analyzer-query-simple',
                    },
                    'speaker': {
                        'type': 'keyword',
                        'ignore_above': 256,
                    },
                },
            },
        },
    },
    'settings': {
        'index': {
            'similarity': {
                'kaldi': {
                    'type': 'kaldi',
                    'mode': 'combined',
                },
            },
        },
        'analysis': {
            'char_filter': {
                'acronym_fixer': {
                    'type': 'pattern_replace',
                    'pattern': '(?<![a-zA-Z])([a-zA-Z])\\. (?=[a-zA-Z]\\.)',
                    'replacement': '$1.',
                },
            },
            'filter': {
                'payload_poslen': {
                    'type': 'payload-position-length',
                },
                'poslen_order': {
                    'type': 'position-length-order',
                },
                'english_stemmer': {
                    'type': 'stemmer',
                    'name': 'minimal_english',
                },
                'unique_stem': {
                    'type': 'unique',
                    'only_on_same_position': True,
                },
            },
            'analyzer': {
                'analyzer-cats-index': {
                    'type': 'custom',
                    'tokenizer': 'tokenizer-cats',
                    'filter': ['lowercase', 'poslen_order', 'payload_poslen'],
                },
                'analyzer-query': {
                    'type': 'custom',
                    'tokenizer': 'standard',
                    'filter': ['lowercase', 'keyword_repeat', 'english_stemmer', 'unique_stem'],
                },
                'analyzer-query-simple': {
                    'type': 'custom',
                    'tokenizer': 'whitespace',
                    'filter': ['lowercase'],
                },
            },
        },
    },
}


class ESWrapper:
    def __init__(
        self,
        hosts=ES_HOSTS,
        port=ES_PORT,
        use_ssl=False,
        verify_certs=False,
        api_key=None,
    ):
        self.logger = logging.getLogger(__name__)
        self.template = TEMPLATE

        # Elastic cloud cluster uses an API key stored on S3 to authenticate requests.
        if not api_key:
            api_key = {'id': 'foo', 'api_key': 'bar'}
        try:
            self.es = Elasticsearch(
                hosts,
                api_key=(api_key['id'], api_key['api_key']),
                port=port,
                use_ssl=use_ssl,
                verify_certs=verify_certs,
                connection_class=RequestsHttpConnection,
            )
        except Exception as e:
            self.logger.error("Failed to connect to Elasticsearch: %s", str(e))
            raise e

    def exists(self, index, doc_id=None):
        """Check if index exists. Checks if document exists if doc_id supplied."""
        if doc_id:
            return self.es.exists(index=index, id=doc_id)
        return self.es.indices.exists(index=index)

    def create(self, index):
        """Create index if necessary."""
        ic = self.es.indices
        if self.exists(index):
            return

        try:
            # Flush to give time for potentially time-consuming index creation.
            ic.create(
                index=index,
                body=self.template,
                request_timeout=ES_INDEX_CREATION_TIMEOUT_SECONDS,
            )
            return ic.flush(
                index=index,
                wait_if_ongoing=True,
                request_timeout=ES_INDEX_CREATION_TIMEOUT_SECONDS,
            )
        except RequestError as e:
            if getattr(e, 'error') == 'resource_already_exists_exception':
                return
            else:
                self.logger.error('Unexpected exception on index creation')
                raise e

    def index(self, index, doc_id, item):
        """Index resource item as document"""
        if not item:
            raise ValueError('You need to set an item to be indexed.')

        try:
            self.create(index)
            self.logger.info("Indexing %s/%s", index, doc_id)
            return self.es.index(index=index, id=doc_id, body=item)
        except Exception as e:
            self.logger.error("elasticsearch.index.index: %s", repr(e))
            raise e


def index_lines(json_lines_by_speaker, es, index, title):
    """Index speaker-separated reply messages passed in as JSON-formatted lines."""
    tracks = []
    for speaker, json_lines in json_lines_by_speaker.items():
        concats = []
        for json_line in json_lines:
            engine_reply = json.loads(json_line)
            if engine_reply.get('final'):
                concats += engine_reply.get('phrases', '')
        if len(concats) > 0:
            tracks.append({'asr': json.dumps(concats), 'speaker': speaker})

    if len(tracks) > 0:
        document = {'title': title, 'tracks': tracks}
        index_id = str(uuid.uuid4())
        es.index(index, index_id, document)
        logging.info("Indexed audio in index %s with ID %s.", index, index_id)
    else:
        logging.info('Not indexing; no "phrases" fields found.')


def get_speaker_names(speakers_string='', number_speaker_tracks=1):
    speakers = [] if speakers_string == '' else speakers_string.split(sep=',')
    if len(speakers) != number_speaker_tracks:
        if number_speaker_tracks == 1:
            speakers = ['Sole Speaker']
        else:
            speakers = [f"Speaker {i}" for i in range(number_speaker_tracks)]
        if speakers_string != '':
            logging.warning(
                "Expected %d speakers input, instead got '%s'. Falling back to %s",
                number_speaker_tracks,
                speakers_string,
                speakers,
            )
    return speakers


def search_index(terms, es, index_name, speakers=None, body=None):
    """Search for given terms, filtering by speakers if given."""
    if body is None:
        phrase_query = {
            'match_payload_phrase': {
                'tracks.asr': {
                    'query': terms,
                },
            },
        }
        and_query = {
            'match': {
                'tracks.asr': {
                    'operator': 'AND',
                    'query': terms,
                },
            },
        }
        or_query = {
            'match': {
                'tracks.asr': {
                    'operator': 'OR',
                    'query': terms,
                },
            },
        }
        if len(terms.split(sep=' ')) == 1:
            if not speakers:
                inner_query = phrase_query
            else:
                inner_query = {'bool': {'must': phrase_query}}
        else:
            inner_query = {
                'bool': {
                    'should': [
                        phrase_query,
                        and_query,
                        or_query,
                    ],
                },
            }
        if speakers:
            speakers_filter = {
                'filter': {'terms': {'tracks.speaker': [speaker for speaker in speakers]}},
                'minimum_should_match': 2,
            }
            inner_query['bool'].update(speakers_filter)
        nested_query = {
            'nested': {
                'path': 'tracks',
                'query': inner_query,
            },
        }
        body = {
            '_source': [
                'title',
                'tracks.speaker',
            ],
            'highlight': {
                'fields': {
                    'tracks.speaker': {},
                    'tracks.asr': {
                        'type': 'highlighter-cats',
                        'pre_tags': ['<em>'],
                        'post_tags': ['</em>'],
                    },
                },
            },
            'query': nested_query,
        }
    logging.info("POST request with query:\n%s", json.dumps(body, indent=2))
    return es.es.search(body=body, index=index_name)


def main():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        'es_host',
        metavar='ES_HOST',
        help='Elasticsearch server host.',
        default='http://localhost',
    )
    parser.add_argument(
        'es_port',
        metavar='ES_PORT',
        help='Elasticsearch server port.',
        default='9200',
    )
    parser.add_argument(
        '--speakers',
        help='Name(s) of speaker(s) to add to index, or to filter for in search.',
        default='',
    )
    parser.add_argument(
        '--index-name',
        help='Name of index to add results to.',
        default='mod9-asr',
    )
    parser.add_argument(
        '--title',
        help='Title to apply to recording in index.',
        default='untitled',
    )
    parser.add_argument(
        '--api-key-id',
        help='API key ID for ES cluster.',
        default=None,
    )
    parser.add_argument(
        '--api-key-secret',
        help='API key secret for ES cluster.',
        default=None,
    )
    parser.add_argument(
        '--log-level',
        metavar='LEVEL',
        help='Verbosity of logging.',
        default='INFO',
    )
    actions = parser.add_mutually_exclusive_group()
    actions.add_argument(
        '--index',
        help='Index from comma-delimited list of files.',
        default='',
    )
    actions.add_argument(
        '--search',
        help='Search given terms. If this flag is not set, index JSON lines from stdin.',
        default='',
    )
    args = parser.parse_args()

    args.log_level = args.log_level.upper()
    logging.basicConfig(
        format="[%(asctime)s] %(message)s",
        level=args.log_level,
    )

    # Use local time with ISO-8601 format for time output.
    logging.Formatter.formatTime = utils.format_log_time

    api_key = None
    if args.api_key_id is not None and args.api_key_secret is not None:
        api_key = {'id': args.api_key_id, 'api_key': args.api_key_secret}

    es = ESWrapper(
        hosts=args.es_host,
        port=args.es_port,
        api_key=api_key,
    )
    es.create(args.index_name)

    if not args.search:
        if args.index == '':
            logging.info('Indexing from stdin.')
            speaker = get_speaker_names(speakers_string=args.speakers)[0]
            index_lines({speaker: sys.stdin}, es, args.index_name, args.title)
        else:
            logging.info("Indexing from user input %s.", args.index)
            speakers = get_speaker_names(
                speakers_string=args.speakers,
                number_speaker_tracks=len(args.index.split(sep=',')),
            )
            json_lines_by_speaker = dict(
                zip(
                    speakers,
                    [open(input_file) for input_file in args.index.split(sep=',')],
                ),
            )

            index_lines(json_lines_by_speaker, es, args.index_name, args.title)

            for input_file in json_lines_by_speaker.values():
                input_file.close()
    else:
        speakers = [] if args.speakers == '' else args.speakers.split(sep=',')
        result = search_index(
            args.search,
            es,
            args.index_name,
            speakers=speakers,
        )
        print(json.dumps(result, indent=2))


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import argparse
import itertools
import json
import os
import shutil
import subprocess
import sys
import tempfile
import urllib.request

DESCRIPTION = """
This specialized tool can be used to score the Switchboard benchmark by suitably formatting and
scoring output from the Mod9 ASR Engine. It reads lines of JSON (i.e. JSONL format) from stdin and
prints a report on stdout, saving files in a work directory.
It can also read results formatted by Google, Amazon, Microsoft, Deepgram, Rev.ai or IBM.
This uses the official NIST SCTK software, which is expected to be installed on the system, and also
requires certain reference data files which might be downloaded. These dependencies are already
installed in the mod9/asr Docker image for convenience.
The Switchboard audio data is needed for meaningful demonstration and could be obtained from the
Linguistic Data Consortium (https://catalog.ldc.upenn.edu/LDC2002S09).
"""

ALTERNATIVES_LEVELS = [
    'word',        # e.g. MBR-derived sausages
    'phrase',      # patent-pending Mod9 data structure
    'transcript',  # i.e. N-best
]

OMITTED_WORDS = [
    # These non-speech words are not transcribed in the reference, and will always hurt the score.
    '[cough]',        # Kaldi recipes
    '[laughter]',     # Kaldi recipes
    '[noise]',        # Kaldi recipes
    '<affirmative>',  # Rev.ai
    '<laugh>',        # Rev.ai
]

REFERENCE_GLM = 'switchboard-benchmark.glm'
REFERENCE_STM = 'switchboard-benchmark.stm'
REFERENCE_URL = 'https://mod9.io/switchboard-benchmark'

SCLITE_OPTS = '-F -D'

SCTK_TOOLS = [
    'csrfilt.sh',
    'rfilter1',
    'sclite',
]

SWITCHBOARD_SPEAKER_IDS = [
    "sw_4390_A",
    "sw_4390_B",
    "sw_4484_A",
    "sw_4484_B",
    "sw_4507_A",
    "sw_4507_B",
    "sw_4520_A",
    "sw_4520_B",
    "sw_4537_A",
    "sw_4537_B",
    "sw_4543_A",
    "sw_4543_B",
    "sw_4547_A",
    "sw_4547_B",
    "sw_4560_A",
    "sw_4560_B",
    "sw_4577_A",
    "sw_4577_B",
    "sw_4580_A",
    "sw_4580_B",
    "sw_4601_A",
    "sw_4601_B",
    "sw_4604_A",
    "sw_4604_B",
    "sw_4683_A",
    "sw_4683_B",
    "sw_4686_A",
    "sw_4686_B",
    "sw_4689_A",
    "sw_4689_B",
    "sw_4694_A",
    "sw_4694_B",
    "sw_4776_A",
    "sw_4776_B",
    "sw_4824_A",
    "sw_4824_B",
    "sw_4854_A",
    "sw_4854_B",
    "sw_4910_A",
    "sw_4910_B",
]

WORKDIR = '/tmp/switchboard-benchmark'

VERBOSE = None  # HACK: use this as a global variable for logging ... see below.


def info(*args, **kwargs):
    if VERBOSE:
        print(*args, **kwargs)


def error(message):
    print(f"ERROR: {message}")
    exit(1)


def run(command, capture_output=True):
    try:
        return subprocess.run(command, capture_output=capture_output, shell=True, check=True)
    except subprocess.SubprocessError:
        error(f"Unable to run '{command}'.")


def install_reference(reference_filename, reference_url):
    if os.path.exists(reference_filename):
        info(f"Found a previously installed reference file: {reference_filename}")
    else:
        info(f"Did not find {reference_filename}; downloading from {reference_url} ...")
        try:
            with urllib.request.urlopen(os.path.join(reference_url, reference_filename)) as resp:
                with tempfile.NamedTemporaryFile(delete=False) as ntf:
                    shutil.copyfileobj(resp, ntf)
                    os.rename(ntf.name, reference_filename)
        except Exception:
            error(f"Unable to download {reference_filename} from {reference_url}.")


def convert_google_json_to_jsonl(json_filename, jsonl_filename):
    """
    Convert Google STT formatted JSON to ASR Engine formatted JSON lines.
    """
    with open(json_filename, 'r', encoding='utf-8') as f:
        response = json.load(f)
        if 'response' in response:
            # When retrieving a LongRunningRecognize "operation", the "response" is nested.
            response = response['response']

    # Assume the first result starts at time zero; see note below.
    start_time = 0.0

    with open(jsonl_filename, 'w', encoding='utf-8') as jsonl_file:
        for result in response['results']:
            reply = {
                'final': True,
                'transcript': result['alternatives'][0]['transcript']
            }

            reply['alternatives'] = [
                {'transcript': a['transcript']} for a in result['alternatives']
            ]

            end_time = float(result['resultEndTime'][:-1])  # strip the "s" suffix

            if 'words' in result['alternatives'][0]:
                words = result['alternatives'][0]['words']
                reply['words'] = [{'word': w['word']} for w in words]
                if words and 'confidence' in words[0]:
                    for i, w in enumerate(words):
                        reply['words'][i]['confidence'] = w['confidence']
                if words and 'startTime' in words[0]:
                    for i, w in enumerate(words):
                        reply['words'][i]['interval'] = [
                            float(w['startTime'][:-1]), float(w['endTime'][:-1])
                        ]

                    # HACK: if enableWordTimestamps was true, then we can infer the start/end times
                    # of the result as start/end times of first/last words in the first alternative.
                    start_time = float(words[0]['startTime'][:-1])
                    end_time = float(words[-1]['endTime'][:-1])

            if 'resultStartTime' in result:
                # This is not officially part of the Google STT REST API response format, but is
                # provided in Mod9's implementation as a compatible superset of functionality.
                # See the note below regarding the implications for NIST SCTK scoring tools.
                start_time = float(result['resultStartTime'][:-1])

            reply['interval'] = [start_time, end_time]

            if 'phrases' in result:
                reply['phrases'] = [
                    {
                        'phrase': p['phrase'],
                        'interval': [float(p['startTime'][:-1]), float(p['endTime'][:-1])],
                        'alternatives': p['alternatives'],
                    } for p in result['phrases']
                ]

            # NOTE: Google does not provide word-level alternatives, but Mod9's extension does.
            if 'words' in result:
                reply['words'] = [
                    {
                        'word': w['word'],
                        'interval': [float(w['startTime'][:-1]), float(w['endTime'][:-1])],
                        'alternatives': w['alternatives'],
                    } for w in result['words']
                ]

            jsonl_file.write(json.dumps(reply) + '\n')

            # Assume the next result's start time is the end time of this result.
            # Google's format does not provide a start time for each result (i.e. utterance).
            # Using the previous result's end time is an unfortunate workaround for the purpose
            # of producing a CTM, particularly if word-level timestamps are not available and are
            # then inferred as uniformly distributed over the utterance-level interval. This is
            # especially problematic for dual-channel telephony in which there are long stretches
            # of silence between speaker turns, during which words are mistimed.
            start_time = end_time


def convert_deepgram_json_to_jsonl(json_filename, jsonl_filename):
    """
    Convert Deepgram formatted JSON to ASR Engine formatted JSON lines.
    """
    response = json.load(open(json_filename, 'r', encoding='utf-8'))
    with open(jsonl_filename, 'w', encoding='utf-8') as jsonl_file:
        result = response['results']['channels'][0]
        reply = {'final': True}
        reply['transcript'] = result['alternatives'][0]['transcript']
        reply['words'] = [{'word': w['word'], 'interval': [w['start'], w['end']]}
                          for w in result['alternatives'][0]['words']]
        jsonl_file.write(json.dumps(reply) + '\n')


def convert_ibm_json_to_jsonl(json_filename, jsonl_filename):
    """
    Convert IBM Watson formatted JSON to ASR Engine formatted JSON lines.
    """
    response = json.load(open(json_filename, 'r', encoding='utf-8'))
    with open(jsonl_filename, 'w', encoding='utf-8') as jsonl_file:
        for result in response['results'][0]['results']:  # awkwardly nested
            reply = {'final': True}
            reply['transcript'] = result['alternatives'][0]['transcript'].strip()
            reply['words'] = [{'word': w, 'interval': [start, end]}
                              for w, start, end, in result['alternatives'][0]['timestamps']]
            jsonl_file.write(json.dumps(reply) + '\n')


def convert_msft_json_to_jsonl(json_filename, jsonl_filename):
    """
    Convert Microsoft formatted JSON to ASR Engine formatted JSON lines.
    """
    def convert_interval(offset_ticks, duration_ticks):
        start_time = offset_ticks / 10_000_000
        return [round(start_time, 2), round(start_time + duration_ticks / 10_000_000, 2)]

    response = json.load(open(json_filename, 'r', encoding='utf-8'))
    with open(jsonl_filename, 'w', encoding='utf-8') as jsonl_file:
        for result in response['recognizedPhrases']:
            reply = {'final': True}
            reply['transcript'] = result['nBest'][0]['lexical']

            reply['alternatives'] = [
                {'transcript': nbest['lexical']} for nbest in result['nBest']
            ]

            reply['interval'] = convert_interval(result['offsetInTicks'], result['durationInTicks'])
            reply['words'] = [{
                'word': w['word'],
                'interval': convert_interval(w['offsetInTicks'], w['durationInTicks'])
            } for w in result['nBest'][0]['words']]
            jsonl_file.write(json.dumps(reply) + '\n')


def convert_amazon_json_to_jsonl(json_filename, jsonl_filename):
    """
    Convert Amazon Transcribe formatted JSON to ASR Engine formatted JSON lines.
    This only handles results with 1-best transcripts, not transcript-level alternatives.
    """
    response = json.load(open(json_filename, 'r', encoding='utf-8'))
    with open(jsonl_filename, 'w', encoding='utf-8') as jsonl_file:
        result = response['results']
        reply = {
            'final': True,
            'transcript': result['transcripts'][0]['transcript']
        }
        reply['words'] = []
        for item in result['items']:
            if item['type'] == 'pronunciation':  # i.e. a "word"
                word = item['alternatives'][0]['content']
                reply['words'].append({
                    'confidence': item['alternatives'][0]['confidence'],
                    'interval': [float(item['start_time']), float(item['end_time'])],
                    'word': word})
        jsonl_file.write(json.dumps(reply) + '\n')


def convert_revai_json_to_jsonl(json_filename, jsonl_filename):
    """
    Convert Rev.ai formatted JSON to ASR Engine formatted JSON lines.
    """
    def multiword(word):
        """NIST SCTK does not like a "word" with spaces in it, so rewrite it with underscores."""
        return word.replace(' ', '_')

    response = json.load(open(json_filename, 'r', encoding='utf-8'))
    with open(jsonl_filename, 'w', encoding='utf-8') as jsonl_file:
        for segment in response['monologues']:
            reply = {'final': True}
            reply['transcript'] = ' '.join([multiword(x['value']) for x in segment['elements']
                                            if x['type'] == 'text'])
            reply['words'] = [{'word': multiword(w['value']), 'interval': [w['ts'], w['end_ts']]}
                              for w in segment['elements'] if w['type'] == 'text']
            jsonl_file.write(json.dumps(reply) + '\n')


def convert_jsonl_to_ctm(
        jsonl_filename, ctm_filename,
        key1, key2,
        alternatives=None,
        alternatives_max=None,
        omitted_words=[],
        split_initialisms=True,
):
    """
    Convert ASR Engine output (JSON lines) to a NIST-formatted CTM file.

    The keys are used to lookup the corresponding reference in the STM file.
    The first key is traditionally the basename of a 2-channel audio file.
    The second key is traditionally a channel identifier, e.g. "A" or "B".

    If the alternatives argument is specified (as "phrase", "transcript", or
    "word"), the resulting file will make use of the CTM format's ability to
    represent alternative hypotheses. This somewhat lesser-known feature of
    the NIST SCTK software is unfortunately a bit buggy, though, and will
    require further post-processing to be handled correctly after filtering
    with the csrfilt.sh tool. If the alternatives_max argument is specified,
    this will limit the number of alternatives considered.

    The list of omitted_words is used as a simple filter to remove words in
    ASR output that is not well matched to the reference conventions. For
    example, it is never helpful to transcribe non-speech noises, since the
    reference does not include these and they will become insertion errors.

    The split_initialisms argument is used to match a convention in the
    reference transcription: for example, Y M C A should be 4 single-letter
    words rather than one single word such as y._m._c._a.
    """
    with open(jsonl_filename, 'r') as jsonl_file, open(ctm_filename, 'w') as ctm_file:
        def write_ctm(key1, key2, begin_time, duration, word, confidence=None):
            # Some ASR systems have a convention of writing initialisms as "a._b._c."
            # But the Switchboard reference treats this as separate words "a", "b", "c".
            if split_initialisms and ('_' in word or (len(word) == 2 and word.endswith('.'))):
                words = word.split('_')
                duration /= len(words)
                for w in words:
                    if len(w) == 2 and w.endswith('.'):
                        w = w[0]
                    write_ctm(key1, key2, begin_time, duration, w, confidence)
                    begin_time += duration
            else:
                if word == '':
                    word = '@'  # Special NIST SCTK convention to indicate null word.
                ctm_file.write(f"{key1} {key2} {begin_time:0.3f} {duration:0.3f} {word}")
                if confidence:
                    ctm_file.write(f" {confidence}")
                ctm_file.write('\n')

        def write_words(key1, key2, begin_time, duration, words):
            duration /= len(words)
            for word in words:
                write_ctm(key1, key2, begin_time, duration, word)
                begin_time += duration

        for line in jsonl_file:
            reply = json.loads(line)
            if not reply.get('final'):
                # Skip replies that do not represent a finalized transcript result.
                continue
            if alternatives is None:
                if 'words' in reply:
                    for word_obj in reply['words']:
                        if 'interval' not in word_obj:
                            # TODO: we could try to infer these from the transcript-level interval.
                            error('Word-level intervals are expected.')
                        begin_time = word_obj['interval'][0]
                        duration = word_obj['interval'][1] - begin_time

                        word = word_obj['word']
                        if not word:
                            # This happens with the legacy Remeeting ASR API (IBM-compatible format)
                            # to represent situations where a top-ranked word alternative is silence.
                            continue

                        if word in omitted_words:
                            continue

                        confidence = word_obj.get('confidence')

                        write_ctm(key1, key2, begin_time, duration, word, confidence)
                else:
                    if 'interval' not in reply:
                        error('Transcript-level intervals are expected.')
                    begin_time = reply['interval'][0]
                    duration = reply['interval'][1] - begin_time
                    words = [w for w in reply['transcript'].split() if w not in omitted_words]
                    if words:
                        write_words(key1, key2, begin_time, duration, words)
            else:
                if alternatives == 'phrase':
                    if 'phrases' not in reply:
                        error('Phrase-level alternatives are expected.')
                    for phrase_obj in reply['phrases']:
                        if 'interval' not in phrase_obj:
                            error('Phrase-level intervals are expected.')
                        begin_time = phrase_obj['interval'][0]
                        duration = phrase_obj['interval'][1] - begin_time

                        if 'alternatives' not in phrase_obj:
                            error('Phrase-level alternatives are expected.')
                        alts = phrase_obj['alternatives'][:alternatives_max]

                        for alt in alts:
                            words = []
                            for w in alt['phrase'].split():
                                if w in omitted_words:
                                    w = '@'  # Special NIST SCTK convention for optional word.
                                words.append(w)
                            alt['phrase'] = ' '.join(words)
                            if alt['phrase'] == '':
                                alt['phrase'] = '@'

                        ctm_file.write(f"{key1} {key2} * * <ALT_BEGIN>\n")
                        for alt in alts[:-1]:
                            words = alt['phrase'].split()
                            write_words(key1, key2, begin_time, duration, words)
                            ctm_file.write(f"{key1} {key2} * * <ALT>\n")
                        words = alts[-1]['phrase'].split()
                        write_words(key1, key2, begin_time, duration, words)
                        ctm_file.write(f"{key1} {key2} * * <ALT_END>\n")
                elif alternatives == 'transcript':
                    if 'interval' not in reply:
                        error('Transcript-level intervals are expected.')
                    begin_time = reply['interval'][0]
                    duration = reply['interval'][1] - begin_time

                    if 'alternatives' not in reply:
                        error('Transcript-level alternatives are expected.')
                    alts = reply['alternatives'][:alternatives_max]

                    for alt in alts:
                        words = []
                        for w in alt['transcript'].split():
                            if w in omitted_words:
                                w = '@'  # Special NIST SCTK convention for optional word.
                            words.append(w)
                        alt['transcript'] = ' '.join(words)
                        if alt['transcript'] == '':
                            alt['transcript'] = '@'

                    ctm_file.write(f"{key1} {key2} * * <ALT_BEGIN>\n")
                    for alt in alts[:-1]:
                        words = alt['transcript'].split()
                        write_words(key1, key2, begin_time, duration, words)
                        ctm_file.write(f"{key1} {key2} * * <ALT>\n")
                    words = alts[-1]['transcript'].split()
                    write_words(key1, key2, begin_time, duration, words)
                    ctm_file.write(f"{key1} {key2} * * <ALT_END>\n")
                elif alternatives == 'word':
                    if 'words' not in reply:
                        error('Word-level alternatives are expected.')
                    for word_obj in reply['words']:
                        if 'interval' not in word_obj:
                            error('Word-level intervals are expected.')
                        begin_time = word_obj['interval'][0]
                        duration = word_obj['interval'][1] - begin_time

                        if 'alternatives' not in word_obj:
                            error('Word-level alternatives are expected.')
                        alts = word_obj['alternatives'][:alternatives_max]

                        for alt in alts:
                            if alt['word'] == '' or alt['word'] in omitted_words:
                                alt['word'] = '@'  # Special NIST SCTK convention for optional word.

                        # NOTE: in theory we should be able to use confidence for word alternatives.
                        # NIST BUG: cannot apply GLM to CTM with alternatives with confidence.
                        # TODO: report this to Jon Fiscus?
                        confidence = None

                        ctm_file.write(f"{key1} {key2} * * <ALT_BEGIN>\n")
                        for alt in alts[:-1]:
                            word = alt['word']
                            write_ctm(key1, key2, begin_time, duration, word, confidence)
                            ctm_file.write(f"{key1} {key2} * * <ALT>\n")
                        word = alts[-1]['word']
                        write_ctm(key1, key2, begin_time, duration, word, confidence)
                        ctm_file.write(f"{key1} {key2} * * <ALT_END>\n")
                else:
                    error('Unexpected alternatives level.')


def expand_alt_section(alt_section, max_expansions=None):
    """Helper function for refilter_ctm."""
    spans = [['']]
    alt_separator_line = ''
    for line in alt_section.strip().split('\n'):
        if '<ALT_BEGIN>' in line:
            spans.append([''])
        elif '<ALT_END>' in line:
            alt_separator_line = line.replace('<ALT_END>', '<ALT>').strip() + '\n'
            spans.append([''])
        elif '<ALT>' in line:
            spans[-1].append('')
        else:
            spans[-1][-1] += line + '\n'
    alts = list(itertools.product(*spans))
    if max_expansions and len(alts) > max_expansions:
        alts = alts[:max_expansions]
    expanded_alt_sections = [''.join(s) for s in alts]
    return alt_separator_line.join(expanded_alt_sections)


def unique_alt_sections(alt_sections):
    """Optimize downstream processing by removing duplicated (and empty) alt sections."""
    # NOTE: retain the order of alternatives so that tiebreakers with the same number of errors will
    # favor alternatives which occur first (i.e. are more likely in an N-best, or reflect the
    # surface form in a GLM expansion). If we wanted to game the scoring metric, we would sort these
    # in decreasing order of length so as to maximize the number of correct words.
    uniq_alt_sections = {}  # Python dicts iterate keys in insertion order.
    for alt_section in alt_sections:
        curr_alt_section = ''
        for line in alt_section.strip().split('\n'):
            key1, key2, begin_time, duration, word = line.split(None, 4)
            if word == '<ALT>':
                # This was an expanded alt section.
                uniq_alt_sections[curr_alt_section] = None
                curr_alt_section = ''
            else:
                curr_alt_section += line+'\n'
        uniq_alt_sections[curr_alt_section] = None
    return uniq_alt_sections.keys()


def refilter_ctm(ctm_in_filename, ctm_out_filename,
                 max_expansions=None, omit_backchannels=False, omit_hesitations=False):
    """
    Post-process a CTM file that was produced by running the NIST SCTK tool
    csrfilt.sh on an original input CTM including hypothesis alternations.

    These become doubly-nested, e.g.:
    sw_4390 A * * <ALT_BEGIN>
    ...1...
    sw_4390 A * * <ALT_BEGIN>
    ...2...
    sw_4390 A * * <ALT>
    ...3...
    sw_4390 A * * <ALT_END>
    ...4...
    sw_4390 A * * <ALT_END>

    The solution is to expand these into singly-nested alternations, as
    the Cartesian product of the alternated sections, e.g.:
    sw_4390 A * * <ALT_BEGIN>
    ...1...
    ...2...
    ...4...
    sw_4390 A * * <ALT>
    ...1...
    ...3...
    ...4...
    sw_4390 A * * <ALT_END>

    Unfortunately, this can create rather long alternations, particularly
    for transcript-level N-best alternatives. This can cause problems for
    the downstream NIST SCTK sclite software which may segfault due to low
    precision of indices (see https://github.com/usnistgov/SCTK/pull/34).
    To mitigate this, set max_expansions. This hack shouldn't be needed for
    word-level or phrase-level alternatives, or if the bugfix PR is merged.
    """
    in_alt = False
    in_nested_alt = False
    alt_sections = None
    alt_begin_line = None
    alt_end_line = None
    alt_separator_line = None

    with open(ctm_in_filename, 'r') as ctm_in_file, open(ctm_out_filename, 'w') as ctm_out_file:
        for line in ctm_in_file:
            if omit_backchannels:
                line = line.replace('%BCACK', '@')
                line = line.replace('%BCNACK', '@')
            if omit_hesitations:
                line = line.replace('%HESITATION', '@')
            if '<ALT_BEGIN>' in line:
                alt_begin_line = line
                if in_nested_alt:
                    error('Unexpected doubly-nested alternative.')
                elif in_alt:
                    in_nested_alt = True
                    alt_sections[-1] += line
                else:
                    in_alt = True
                    alt_sections = ['']
                continue
            elif '<ALT_END>' in line:
                alt_end_line = line
                alt_separator_line = alt_end_line.replace('<ALT_END>', '<ALT>').strip() + '\n'
                if in_nested_alt:
                    in_nested_alt = False
                    alt_sections[-1] += line
                elif in_alt:
                    in_alt = False
                    ctm_out_file.write(alt_begin_line)
                    alt_sections = [expand_alt_section(a, max_expansions) for a in alt_sections]
                    alt_sections = unique_alt_sections(alt_sections)
                    ctm_out_file.write(alt_separator_line.join(alt_sections))
                    ctm_out_file.write(alt_end_line)
                else:
                    error('Unexpected end of alternative.')
                continue
            elif '<ALT>' in line:
                if in_nested_alt:
                    alt_sections[-1] += line
                elif in_alt:
                    alt_sections.append('')
                continue

            if in_alt or in_nested_alt:
                alt_sections[-1] += line
            else:
                ctm_out_file.write(line)


def refilter_stm(stm_in_filename, stm_out_filename, optional_backchannels=False):
    """
    There is a bug in NIST SCTK w.r.t. placement of parentheses around curly braces and backslashes,
    i.e. when representing an optionally deleteable alternation.
    Also allow for optionally deletable backchannel mappings (i.e. %BCACK and %BCNACK in the GLM).
    """
    if optional_backchannels:
        optional_backchannels_filter = '| sed "s,%BCACK,(%BCACK),g; s,%BCNACK,(%BCNACK),g"'
    else:
        optional_backchannels_filter = ''
    run('sed "s,(/),/,g; s,({,{(,g; s,}),)},g"'
        f" < {stm_in_filename} {optional_backchannels_filter} > {stm_out_filename}")


def convert_stm_1seg(old_stm_filename, new_stm_filename):
    """
    Convert the STM into a single long segment, which can in some cases
    slightly improve the sclite alignment algorithm or minimize WER, e.g.
    in situations where a pair of insertion/deletion errors are across a
    segment boundary and could be consolidated as a single substitution.
    This also helps in situations where the reference segmentation is
    not used by the ASR system (i.e. a more fair evaluation than typical
    for academic research scenarios); slight differences in word-level
    timing may exist between the ASR system and the reference segments.
    """
    curr_key = None  # This function assumes the STM is speaker-specific.
    transcripts = []
    for line in open(old_stm_filename):
        if line.startswith(';;'):  # comment lines
            continue
        key1, key2, spkid, begin_time, end_time, labels, transcript = line.strip().split(None, 6)
        key = (key1, key2, spkid, labels)
        if curr_key and curr_key != key:
            # TODO: it's not that hard to generalize this to a multi-speaker STM file.
            error('Cannot collapse segmentation of a reference STM with multiple speakers.')
        else:
            curr_key = key
            transcripts.append(transcript)
    with open(new_stm_filename, 'w') as f:
        key1, key2, spkid, labels = curr_key
        f.write(f"{key1} {key2} {spkid} 0 999999999 {labels} {' '.join(transcripts)}\n")


def nist_report(switchboard_speaker_id, show_errors=False):
    """
    Read the .ctm, .pra, and .raw files produced by the NIST SCTK sclite tool.
    Produce a less verbose and somewhat more informative report than SCTK would.
    """
    if show_errors:
        errors = []
        print('\nSegments where reference and hypothesis are mis-aligned:')
        for line in open(f"{switchboard_speaker_id}.nist.pra"):
            if line.startswith('Scores:'):
                fields = line.strip().split()
                n_cor = int(fields[5])
                n_sub = int(fields[6])
                n_del = int(fields[7])
                n_ins = int(fields[8])
                n_err = n_sub + n_del + n_ins
                n_ref = n_err + n_cor
                wer = n_err / n_ref * 100
            elif line.startswith('REF:'):
                ref_line = line
            elif line.startswith('HYP:'):
                if n_err > 0:
                    errors.append((n_err, wer, ref_line, line))
        errors.sort(key=lambda x: x[0])  # or x[1] to sort by WER.
        for n_err, wer, ref_line, hyp_line in errors:
            print(ref_line, end='')
            print(hyp_line, end='')
        print()

    # Report the size of the CTM file, compressed.
    run(f"gzip -c {switchboard_speaker_id}.ctm > {switchboard_speaker_id}.ctm.gz")
    ctm_gz_kb = os.path.getsize(f"{switchboard_speaker_id}.ctm.gz") // 1000
    info(f"Size of *.ctm.gz:  {ctm_gz_kb} KB")

    # Report some statistics about the distribution of alternatives.
    depths = []
    widths = []
    curr_depth = 0
    curr_width = 0
    for line in open(f"{switchboard_speaker_id}.ctm"):
        if '<ALT_BEGIN>' in line:
            curr_depth += 1
        elif '<ALT>' in line:
            curr_depth += 1
            widths.append(curr_width)
            curr_width = 0
        elif '<ALT_END>' in line:
            depths.append(curr_depth)
            widths.append(curr_width)
            curr_depth = curr_width = 0
        else:
            curr_width += 1
    if depths and widths:
        N_max = max(depths)   # When requesting N-best, in practice the depth might be less than N.
        N_dec = sorted(depths)[-len(depths)//10]  # The top decile is more informative than the max.
        N_med = sorted(depths)[len(depths)//2]    # The median is more informative than the mean.
        S = len(depths)       # Number of spans/segments.
        W_avg = sum(widths) / len(widths)  # This relates to the average-case performance.
        W_max = max(widths)   # This relates to the worst-case performance, e.g. for sclite scoring.
        info('Alternatives:      '
             f"N_max={N_max} N_dec={N_dec} N_med={N_med} S={S} W_avg={W_avg:0.1f} W_max={W_max}")

    sum_line = run(f"grep Sum {switchboard_speaker_id}.nist.raw").stdout.decode()
    fields = sum_line.strip().split()
    n_ref = int(fields[4])
    n_cor = int(fields[6])
    n_sub = int(fields[7])
    n_del = int(fields[8])
    n_ins = int(fields[9])
    info(f"Sum (#C #S #D #I): {n_cor} {n_sub} {n_del} {n_ins}")

    # These are not standard metrics, but can be rather enlightening.
    precision = n_cor / (n_cor + n_sub + n_ins)
    recall = n_cor / (n_cor + n_sub + n_del)
    info(f"Precision/Recall:  {precision:0.3f} / {recall:0.3f}")

    # This is the standard ASR metric, which has a lot of drawbacks.
    wer = (n_sub + n_del + n_ins) / n_ref * 100
    print(f"{f'[{switchboard_speaker_id}]':<11} WER:  {f'{wer:.2f}':>6}%", flush=True)


def main():
    try:
        main_helper()
        return 0
    except KeyboardInterrupt:
        return 130


def main_helper():
    parser = argparse.ArgumentParser(description=DESCRIPTION,
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        'switchboard_speaker_id',
        metavar='SW_SPEAKER_ID',
        nargs='?',
        help='Switchboard speaker identifier, including channel, such as "sw_4390_A" for example.',
        choices=SWITCHBOARD_SPEAKER_IDS,
    )
    parser.add_argument(
        '--alternatives',
        metavar='LEVEL',
        help='Word-, phrase-, or transcript-level alternatives used for oracle scoring.',
        choices=ALTERNATIVES_LEVELS,
    )
    parser.add_argument(
        '--alternatives-n',
        metavar='INT',
        type=int,
        help='Limit the number of alternatives (i.e. N-best) to be considered.',
    )
    parser.add_argument(
        '--max-expansions',
        metavar='INT',
        type=int,
        help='Mitigate SCTK bug by limiting nested alternative expansions, e.g. transcript-level.',
        default=0,
    )
    parser.add_argument(
        '--omit-hesitations',
        action='store_true',
        help='CTM should not include hesitations (optionally deletable in the STM).',
    )
    parser.add_argument(
        '--omit-backchannels',
        action='store_true',
        help='CTM should not include backchannels (cf. --optional-backchannels).',
    )
    parser.add_argument(
        '--omitted-words',
        metavar='LIST',
        default=','.join(OMITTED_WORDS),
        help='Comma-separated list of non-speech words to omit from the CTM.',
    )
    parser.add_argument(
        '--optional-backchannels',
        action='store_true',
        help='Backchannels can be optionally deleted in the STM.',
    )
    parser.add_argument(
        '--reference-glm',
        metavar='FILE',
        help='Name of NIST-formatted Global Language Mapping file.',
        default=REFERENCE_GLM,
    )
    parser.add_argument(
        '--reference-stm',
        metavar='FILE',
        help='Name of NIST-formatted Segment Time Mark file.',
        default=REFERENCE_STM,
    )
    parser.add_argument(
        '--reference-url',
        metavar='URL',
        help='Whence missing reference files may be downloaded.',
        default=REFERENCE_URL,
    )
    parser.add_argument(
        '--sclite',
        action='store_true',
        help='Run NIST sclite tool over all files.',
    )
    parser.add_argument(
        '--show-errors',
        action='store_true',
        help='Compare segment-level mis-alignment.',
    )
    parser.add_argument(
        '--single-segment-stm',
        action='store_true',
        help='Convert reference to a long segment.',
    )
    parser.add_argument(
        '--sum-overall',
        action='store_true',
        help='Report aggregate scores over corpus.',
    )
    parser.add_argument(
        '--verbose', '-v',
        action='store_true',
        help='Print more verbose information.',
    )
    parser.add_argument(
        '--workdir',
        metavar='DIRECTORY',
        help='Where files will be saved for caching or debugging.',
        default=WORKDIR,
    )
    args = parser.parse_args()

    global VERBOSE
    VERBOSE = args.verbose

    # This tool will report on stdout, but these saved files might be helpful for debugging.
    os.makedirs(args.workdir, exist_ok=True)
    os.chdir(args.workdir)

    if args.sclite:
        run('cat *.ctm.refilt > overall.ctm.refilt_')
        run('cat *.stm.refilt > overall.stm.refilt_')
        run('sclite -h overall.ctm.refilt_ ctm -r overall.stm.refilt_ stm -n overall.nist '
            f"{SCLITE_OPTS} -o sum",
            capture_output=False)
        run("cat overall.nist.sys", capture_output=False)
        exit(0)
    elif args.sum_overall:
        run("""cat sw_*.nist.raw | grep Sum | awk '
{Snt+=$4;Wrd+=$5;Corr+=$7;Sub+=$8;Del+=$9;Ins+=$10} END \
{print" | Sum | "Snt"  "Wrd" | "Corr" "Sub" "Del" "Ins} \
' > overall.nist.raw""")
        run('cat *.ctm > overall.ctm')
        nist_report('overall')
        exit(0)
    elif not args.switchboard_speaker_id:
        error('Missing required arugment for Switchboard speaker ID.')

    info(f"Results will be saved in the work directory: {args.workdir}")

    # Check installed dependencies.
    install_reference(args.reference_glm, args.reference_url)
    install_reference(args.reference_stm, args.reference_url)
    for sctk_tool in SCTK_TOOLS:
        if not shutil.which(sctk_tool):
            error(f"Could not find {sctk_tool}; ensure that NIST SCTK is installed.")

    # Parse the Switchboard speaker identifier, renamed as `spkid` for convenience.
    spkid = args.switchboard_speaker_id
    filename_id, channel_id = spkid.rsplit('_', 1)

    lines = []
    info("Read Engine replies or other vendors' JSON on stdin: ...", flush=True)
    for line in sys.stdin:
        lines.append(line)

    if lines and lines[0].startswith('{"metadata"'):
        info(f"Save JSON (Deepgram formatted) from stdin: {spkid}.json")
        with open(spkid+'.json', 'w') as f:
            for line in lines:
                f.write(line)
        info(f"Convert JSON to Engine formatted JSON lines: {spkid}.jsonl")
        convert_deepgram_json_to_jsonl(spkid+'.json', spkid+'.jsonl')
    elif lines and lines[0] == '{\n' and ('"name"' in lines[1] or '"results"' in lines[1]):
        info(f"Save JSON (Google STT formatted) from stdin: {spkid}.json")
        with open(spkid+'.json', 'w') as f:
            for line in lines:
                f.write(line)
        info(f"Convert JSON to Engine formatted JSON lines: {spkid}.jsonl")
        convert_google_json_to_jsonl(spkid+'.json', spkid+'.jsonl')
    elif lines and lines[0] == '{\n' and '"created"' in lines[1]:
        info(f"Save JSON (IBM Watson formatted) from stdin: {spkid}.json")
        with open(spkid+'.json', 'w') as f:
            for line in lines:
                f.write(line)
        info(f"Convert JSON to Engine formatted JSON lines: {spkid}.jsonl")
        convert_ibm_json_to_jsonl(spkid+'.json', spkid+'.jsonl')
    elif lines and lines[0] == '{\n' and '"source"' in lines[1]:
        info(f"Save JSON (Microsoft formatted) from stdin: {spkid}.json")
        with open(spkid+'.json', 'w') as f:
            for line in lines:
                f.write(line)
        info(f"Convert JSON to Engine formatted JSON lines: {spkid}.jsonl")
        convert_msft_json_to_jsonl(spkid+'.json', spkid+'.jsonl')
    elif lines and lines[0].startswith('{"jobName"'):
        info(f"Save JSON (as Amazon Transcribe) from stdin: {spkid}.json")
        with open(spkid+'.json', 'w') as f:
            for line in lines:
                f.write(line)
        info(f"Convert JSON to Engine formatted JSON lines: {spkid}.jsonl")
        convert_amazon_json_to_jsonl(spkid+'.json', spkid+'.jsonl')
    elif lines and lines[0].startswith('{"monologues"'):
        info(f"Save JSON (as Rev.ai) from stdin: {spkid}.json")
        with open(spkid+'.json', 'w') as f:
            for line in lines:
                f.write(line)
        info(f"Convert JSON to Engine formatted JSON lines: {spkid}.jsonl")
        convert_revai_json_to_jsonl(spkid+'.json', spkid+'.jsonl')
    else:
        info(f"Save JSON lines (Engine replies) from stdin: {spkid}.jsonl")
        with open(spkid+'.jsonl', 'w') as f:
            for line in lines:
                f.write(line)

    info(f"Convert to NIST-formatted hypothesis format: {spkid}.ctm")
    convert_jsonl_to_ctm(spkid+'.jsonl', spkid+'.ctm',
                         filename_id, channel_id,
                         alternatives=args.alternatives if args.alternatives_n != 0 else None,
                         alternatives_max=args.alternatives_n,
                         omitted_words=args.omitted_words.split(','))

    info(f"Apply global mappings to make filtered file: {spkid}.ctm.filt")
    run(f"csrfilt.sh -i ctm -t hyp -dh {args.reference_glm} < {spkid}.ctm > {spkid}.ctm.filt")

    info(f"Fix SCTK bug, omit hesitations/backchannels: {spkid}.ctm.refilt")
    refilter_ctm(spkid+'.ctm.filt', spkid+'.ctm.refilt',
                 args.max_expansions, args.omit_backchannels, args.omit_hesitations)

    info(f"Extract reference transcription for speaker: {spkid}.stm")
    run(f"grep '^{filename_id} {channel_id} {spkid}' < {args.reference_stm} > {spkid}.stm")

    info(f"Apply global mappings to make filtered file: {spkid}.stm.filt")
    run(f"csrfilt.sh -i stm -t ref -dh {args.reference_glm} < {spkid}.stm > {spkid}.stm.filt")

    info(f"Fix SCTK bugs (optional alts, backchannels): {spkid}.stm.refilt")
    refilter_stm(spkid+'.stm.filt', spkid+'.stm.refilt', args.optional_backchannels)

    stm = f"{spkid}.stm.refilt"
    if args.single_segment_stm:
        info(f"Convert the reference into one long segment: {spkid}.stm.refilt.1seg")
        convert_stm_1seg(stm, f"{spkid}.stm.refilt.1seg")
        stm = f"{spkid}.stm.refilt.1seg"

    info(f"Run the NIST SCLITE tool to produce reports: {spkid}.nist.*", flush=True)
    run(f"sclite -h {spkid}.ctm.refilt ctm -r {stm} stm -n {spkid}.nist {SCLITE_OPTS}"
        " -o pralign rsum")

    nist_report(spkid, args.show_errors)


if __name__ == '__main__':
    main()



"""
Python SDK wrappers over the Mod9 ASR Engine TCP Server.

Requires a Mod9 ASR Engine TCP server backend to work. Contact
sales@mod9.com for licensing inquiries.

More documentation and example usage can be found at
http://mod9.io/python-sdk


``mod9.asr.speech`` is referred to as the Mod9 ASR Python SDK wrapper,
and is a fully-compatible drop-in replacement for the Google STT Python
Client Library, making use of Google's objects. Send input in Google-
style objects or dicts as you would to the Google STT Python Client
Library and the request will be converted to Mod9-compatible form. The
results from the Mod9 ASR Engine TCP server will be converted to Google-
type objects.


``mod9.asr.speech_mod9`` extends the Mod9 ASR Python SDK wrapper with
Mod9-exclusive options, such as phrase alternatives and A-law audio
encodings (in addition to Google offerings such as transcript
alternatives/N-best lists and 16-but linear PCM and μ-law encodings).
Input can use either Google-type or Google-like subclasses within the
module (required to access Mod9-only options) or dict. Output is Google-
like subclasses defined within the module.
"""
from mod9.reformat.config import WRAPPER_VERSION as __version__



#!/usr/bin/env python3

"""
Mod9 ASR WebSocket server: wrapper over the ASR Engine.

This application-level protocol over WebSocket is similar to the Engine's
custom application-level protocol over TCP, but leverages message events.

The first message should be sent by the client as a JSON-formatted object
representing the request options, as in the protocol over TCP. Instead of
strictly requiring single-line newline-terminated JSON, however, clients
of the WebSocket interface may format the JSON with arbitrary whitespace.

Subsequent messages sent by the client should be bytes of the audio data,
as in the protocol over TCP. Each of these messages must be non-empty.

Upon receipt of an empty message, this WebSocket server will consider a
client to have finished sending its audio data. This functionality is in
addition to the underlying Engine's potential to terminate requests as
per the custom application-level protocol over TCP, i.e. after:
 - receiving the number of bytes set by a ``content-length`` option;
 - receiving the number of samples indicated in a WAV header;
 - receiving an "END-OF-FILE" sequence (cf. ``eof`` option);
 - or if no data has been received for over 10 seconds.

The WebSocket client asynchronously receives a series of messages relayed
from the Engine, each encoding a single JSON object. This is similar to
the Engine's response structure for the protocol over TCP, except that no
newlines are required to terminate each reply.

Multiple clients can be served concurrently by this WebSocket server, but
beware that its scalability and robustness is somewhat uncertain:
See https://websockets.readthedocs.io/en/stable/topics/deployment.html
"""

import argparse
import asyncio
import json
import logging
import uuid
import sys

import websockets

import mod9.reformat.config as config
from mod9.reformat import utils

DEBUG_LOGGING_MESSAGE_INTERVAL = 10000

# Incrementing request ID counter and logger adapter to prepend to each request.
request_id = 1


class CustomLoggerAdapter(logging.LoggerAdapter):
    """
    Add ``request_id`` and ``rest_uuid`` to logs.

    E.g. see:
    https://docs.python.org/3/howto/logging-cookbook.html#using-loggeradapters-to-impart-contextual-information
    https://github.com/python/cpython/blob/96cf5a63d2dbadaebf236362b4c7c09c51fda55c/Lib/logging/__init__.py#L1799-L1883
    """
    def process(self, msg, kwargs):
        """Prepend [{ws_uuid}:{request_id}:{level}] to message."""
        ws_uuid = self.extra.get('ws_uuid', str(uuid.UUID(int=0)))[:8]
        request_id = self.extra.get('request_id', 0)
        level = logging.getLevelName(self.extra['level']).lower()
        return f"[{ws_uuid}:{request_id}:{level}]\t{msg}", kwargs

    def log(self, level, msg, *args, **kwargs):
        """Override: pass level as element of ``self.extra`` dict."""
        if self.isEnabledFor(level):
            self.extra['level'] = level
            msg, kwargs = self.process(msg, kwargs)
            self.logger.log(level, msg, *args, **kwargs)


async def relay_options(websocket, sock_writer, logger):
    """
    Relay options from client's (first) WebSocket message to Engine TCP socket.

    Args:
        websocket (websockets.server.WebSocketServerProtocol):
            Client WebSocket to receive options from.
        sock_writer (asyncio.StreamWriter):
            Engine TCP socket to write options to.

    Returns:
        bytes:
            End-of-file byte sequence, possibly set by client options.

    Raises:
        utils.Mod9BadRequestError:
            If client's first message cannot be parsed properly.
    """
    options_message = await websocket.recv()
    logger.debug("Received options message: %s", options_message)

    # Re-serialize WebSocket-specified options as single-line JSON terminated with a newline.
    try:
        options_json = options_message.decode()
        options = json.loads(options_json)
        options_json = json.dumps(options, separators=(',', ':'))
        options_line = options_json.encode('utf-8') + b'\n'
    except json.decoder.JSONDecodeError as e:
        raise utils.Mod9BadRequestError(
            f"Could not parse options from first message (JSON decode error: {e})."
        )
    except Exception:
        raise utils.Mod9BadRequestError(
            f"Could not parse options from first message (comprising {len(options_message)} bytes)."
        )

    # Write the single-line newline-terminated options to Engine over TCP.
    sock_writer.write(options_line)
    await sock_writer.drain()

    # The WebSocket server will need this later in order to properly terminate the request.
    eof = options.get('eof', 'END-OF-FILE').encode('utf-8')
    return eof


async def receive_messages(websocket, logger):
    """
    Receive messages from WebSocket and yield as async bytes generator.
    Stop iteration upon receipt of an empty message.

    Args:
        websocket (websockets.server.WebSocketServerProtocol):
            WebSocket to read from.

    Yields:
        Iterable[bytes]
    """
    num_messages = 0
    total_bytes = 0

    while True:
        message = await websocket.recv()
        num_messages += 1
        total_bytes += len(message)

        if message:
            if num_messages % DEBUG_LOGGING_MESSAGE_INTERVAL == 0:
                logger.debug(
                    "Received %d messages totaling %d bytes.",
                    num_messages,
                    total_bytes,
                )
            yield message
        else:
            # Received empty message, indicating end of messages.
            logger.debug("Received %d messages totaling %d bytes.", num_messages, total_bytes)
            return


async def write_chunks_to_socket(async_bytes_generator, sock_writer, eof):
    """
    Write contents of async bytes generator to given socket.
    The bytes from the generator might be re-chunked.

    Args:
        async_bytes_generator (Iterable[bytes]):
            Data to send to socket.
        sock_writer (asyncio.StreamWriter):
            Socket to send to.
        eof (bytes):
            End-of-file byte sequence.

    Returns:
        None
    """
    async for chunk in async_bytes_generator:
        for i in range(0, len(chunk), config.MAX_CHUNK_SIZE):
            subchunk = chunk[i:i+config.MAX_CHUNK_SIZE]
            sock_writer.write(subchunk)
            await sock_writer.drain()

    # This may not be necessary if the client sent WAV, or already sent their own EOF,
    #  but it seems the Engine doesn't mind receiving an extra EOF in those situations.
    sock_writer.write(eof)
    await sock_writer.drain()


async def read_lines_from_socket(sock_reader):
    """
    Read lines from socket and yield as async lines generator.

    Args:
        sock_reader (asyncio.StreamReader):
            Socket to read lines from.

    Yields:
        Iterable[bytes]
    """
    while True:
        try:
            line = await sock_reader.readline()
        except ValueError:
            raise utils.Mod9LimitOverrunError(
                'Could not read in reply from Engine due to current limit of'
                f" {config.WEBSOCKET_LIMIT_BYTES} bytes."
                ' Please retry the request with a larger value passed to the'
                ' --websocket-limit-bytes command-line argument.'
            )
        if not line:
            return
        yield line


async def send_messages(async_lines_generator, websocket, logger):
    """
    Send contents of async lines generator to given WebSocket.
    Each line is stripped before being sent as a message.

    Args:
        async_lines_generator (Iterable[bytes]):
            Lines to send to socket.
        websocket (websockets.server.WebSocketServerProtocol):
            WebSocket to send to.

    Returns:
        None
    """
    first_response_line = True
    async for line in async_lines_generator:
        message = line.decode().strip().encode('utf-8')
        await websocket.send(message)
        logger.debug("Sent message: %s", message)

        if first_response_line:
            # To pass through Engine UUID in logging.
            engine_uuid = json.loads(message.decode()).get('uuid')
            engine_version = json.loads(message.decode()).get('engine', {}).get('version')
            logger.info(
                "Engine version %s and UUID %s received from first Engine reply.",
                'missing' if not engine_version else engine_version,
                'not' if not engine_uuid else engine_uuid,
            )
            first_response_line = False


async def send_error(websocket, details, logger):
    """
    Send an error message to the WebSocket client, formatted like Engine reply.

    Args:
        websocket (websockets.server.WebSocketServerProtocol):
            WebSocket to send to.
        details (str):
            Information to be included in the .error field, to be prefixed with
            a [WebSocket] tag clarifying that this didn't come from the Engine.

    Returns:
        None
    """
    reply = {'status': 'failed', 'error': '[WebSocket] '+details}
    try:
        reply_json = json.dumps(reply, separators=(',', ':'))
        reply_message = reply_json.encode('utf-8')
        await websocket.send(reply_message)
    except Exception:
        logger.error('Could not send exception info to client.', exc_info=True)
    else:
        logger.debug("Sent error message: %s", reply_message)


async def handle_request(websocket, path):
    """
    Asynchronously send options and audio from WebSocket to Engine
    socket, then send response from Engine socket to WebSocket.

    Args:
        websocket (websockets.server.WebSocketServerProtocol):
            WebSocket from client request.
        path (string):
            WebSocket URI path from client request.

    Returns:
        None
    """
    global request_id
    ws_uuid = str(uuid.uuid4())
    logger = CustomLoggerAdapter(logging.root, {'ws_uuid': ws_uuid, 'request_id': request_id})
    request_id += 1

    logger.info(
        "Request from client %s at path %s assigned WS UUID %s.",
        websocket.remote_address[0],
        path,
        ws_uuid,
    )

    # Resources that might be created below, and should then be cleaned up later.
    sock_writer, websocket_to_socket = None, None

    try:
        # Connect to the ASR Engine over TCP.
        sock_reader, sock_writer = await asyncio.open_connection(
            host=config.ASR_ENGINE_HOST,
            port=config.ASR_ENGINE_PORT,
            limit=config.WEBSOCKET_LIMIT_BYTES,
        )
    except Exception:
        logger.error('Could not connect to ASR Engine.')
        await send_error(
            websocket,
            'Could not connect to ASR Engine; contact server operator.',
            logger,
        )
        return

    try:
        # Send the request options, and parse the EOF byte sequence.
        eof = await relay_options(websocket, sock_writer, logger)

        # Receive messages from WebSocket & write chunks (i.e. audio) to Engine socket.
        websocket_to_socket = asyncio.get_event_loop().create_task(
            write_chunks_to_socket(receive_messages(websocket, logger), sock_writer, eof),
        )

        # Read lines (i.e. replies) from Engine socket & send messages to WebSocket.
        socket_to_websocket = asyncio.get_event_loop().create_task(
            send_messages(read_lines_from_socket(sock_reader), websocket, logger),
        )

        # Wait until the Engine closes the socket.
        # TODO: it might be nice if we inspected and logged the .status of the final Engine message.
        await socket_to_websocket
        logger.debug('ASR Engine closed TCP connection.')
    except utils.Mod9BadRequestError as e:
        logger.error('Request failed due to bad request from client.')
        await send_error(websocket, str(e), logger)
    except websockets.ConnectionClosed:
        logger.error('WebSocket closed unexpectedly; client disconnected?')
        # We cannot send_error to the client in this case, since the WebSocket is closed.
    except Exception:
        # Include the traceback so operators can contact support@mod9.com with helpful information.
        logger.error('Request failed unexpectedly.', exc_info=True)
        # Do not relay details to the client, as it's likely too server-specific or sensitive.
        await send_error(
            websocket,
            'Request failed unexpectedly; contact server operator.',
            logger,
        )
    finally:
        # Cancel any audio remaining to be sent, since the Engine can't accept it anymore.
        if websocket_to_socket:
            websocket_to_socket.cancel()

        # Close the underlying socket; this is not a TCP half-close (cf. write_eof).
        if sock_writer:
            sock_writer.close()
            await sock_writer.wait_closed()

        # Drain messages left over in WebSocket recv buffer to ensure reply CLOSE frame is received,
        #  e.g. if the client has sent more data than the Engine was able to accept.
        websocket_close_task = asyncio.get_event_loop().create_task(websocket.close())
        try:
            while True:
                _ = await websocket.recv()  # Discard received message.
        except websockets.ConnectionClosedOK:
            logger.debug('WebSocket connection drained successfully.')
        except Exception:
            logger.warning('WebSocket connection was not drained successfully.')

        await websocket_close_task
        logger.info("WebSocket connection closed.")


def main():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        '--engine-host',
        metavar='HOST',
        help='ASR Engine host name.'
             ' Can also be set by ASR_ENGINE_HOST environment variable.',
        default=config.ASR_ENGINE_HOST,
    )
    parser.add_argument(
        '--engine-port',
        metavar='PORT',
        help='ASR Engine port number.'
             ' Can also be set by ASR_ENGINE_PORT environment variable.',
        type=int,
        default=config.ASR_ENGINE_PORT,
    )
    parser.add_argument(
        '--host',
        help='WebSocket host address. Can be set to 0.0.0.0 for external access.',
        default='127.0.0.1'  # Internal access only.
    )
    parser.add_argument(
        '--port',
        help='WebSocket port number.',
        type=int,
        default=9980,  # The ASR Engine is typically at 9900, so add 80 (i.e. standard HTTP port).
        # NOTE: the default for the REST API should be 8080, so try to avoid a conflict with that.
    )
    parser.add_argument(
        '--log-level',
        metavar='LEVEL',
        help='Verbosity of logging.',
        default='INFO',
    )
    parser.add_argument(
        '--skip-engine-check',
        action='store_true',
        help='When starting server, do not wait for ASR Engine.',
        default=False,
    )
    parser.add_argument(
        '--websocket-limit-bytes',
        help='Limit on number of bytes allowed per reply line read from Engine. Default: 1 MiB.',
        type=int,
    )
    args = parser.parse_args()

    # TODO: we should instead adjust the log level of our own logger rather than the root logger.
    # TODO: Trace request IDs in logging.
    args.log_level = args.log_level.upper()
    logging.basicConfig(
        format="[%(asctime)s] %(message)s",
        level=args.log_level,
    )

    # Use local time with ISO-8601 format for time output.
    logging.Formatter.formatTime = utils.format_log_time

    if args.log_level == 'DEBUG':
        # These are overly verbose for our purposes.
        logging.getLogger('asyncio').setLevel(logging.INFO)
        logging.getLogger('websockets').setLevel(logging.INFO)

    config.ASR_ENGINE_HOST = args.engine_host
    config.ASR_ENGINE_PORT = args.engine_port

    if args.websocket_limit_bytes:
        config.WEBSOCKET_LIMIT_BYTES = args.websocket_limit_bytes

    logger = CustomLoggerAdapter(logging.root, dict())

    if not args.skip_engine_check:
        utils.test_host_port(logger=logger)

    logger.info(
        "Running a WebSocket server (Python SDK version %s) at %s on port %d.",
        config.WRAPPER_VERSION,
        args.host,
        args.port,
    )
    start_server = websockets.serve(handle_request, host=args.host, port=args.port)
    asyncio.get_event_loop().run_until_complete(start_server)

    try:
        asyncio.get_event_loop().run_forever()
    except KeyboardInterrupt:
        logger.warning('Exiting on interrupt signal.')
        sys.exit(130)  # Don't show Python traceback.


if __name__ == '__main__':
    main()



"""
Provides defaults used throughout mod9-asr.
"""

import logging
import os

# Current wrappers version.  Note that this is not the same as the Engine version.
WRAPPER_VERSION = '1.11.4'

# CHANGELOG:
#   1.11.4 (13 May 2023):
#   - Pin Flask to version <2.3 due to deprecations.
#   1.11.3 (19 May 2022):
#   - Fix major bugs in Python SDK caused by an internal refactoring circa version 1.7.0.
#   1.11.2 (28 Mar 2022):
#   - Fix minor bug to allow multiple rescoring after running with --sclite option.
#   1.11.1 (22 Mar 2022):
#   - Fix minor bug to allow asrModel request option to be specified.
#   1.11.0 (21 Mar 2022):
#   - Allow longer connection timeout, useful for benchmarking with speed=3.
#   - Support multichannel audio for any number of channels, unlike Google that has restrictions
#     based on the encoding. Internally, the Engine supports a max of 128 byte-aligned samples.
#     for 16-bit encoding, that is 64 channels.
#     - The request option "audioChannelCount" indicates the number of channels.
#     - The request option "enableSeparateRecognitionPerChannel" must be `true` for multichannel
#       audio recognized via Mod9. This is unlike Google that recognizes only the first channel
#       for the default value of `false`.
#     - The response field "channelTag" indicates the channel number of the segment.
#   1.10.0 (16 Feb 2022):
#   - Ignore empty word in Switchboard benchmark scoring script (for legacy Remeeting ASR API).
#   - Add Rev.ai formatted result handling to Switchboard benchmark script.
#   1.9.0 (09 Feb 2022):
#   - Add Deepgram formatted result handling to Switchboard benchmark script.
#   1.8.0 (08 Feb 2022):
#   - Add Microsoft formatted result handling to Switchboard benchmark script.
#   1.7.0 (02 Feb 2022):
#   - Allow multiple clients to connect to different Engine hosts and ports.
#   - Support more audio encodings: 24- and 32-bit signed integers and 32-bit float.
#   - Add IBM Watson format conversion to mod9-asr-switchboard-benchmark.
#   1.6.0 (10 Jan 2022):
#   - Add "asrModel" to results using speech_mod9 and REST API.
#   - Minor fixes to mod9-asr-switchboard-benchmark and new features:
#     - Enable scoring of 1-best results in Amazon Transcribe formatted JSON.
#     - Set --optional-backchannels to make %BCACK and %BCNACK optionally deletable in the STM.
#     - Set --omit-backchannels to remove from the CTM, since it won't help if optionally deletable.
#     - Set --omit-hesitations to remove %HESITATION from the CTM, since it's optionally deletable.
#     - Change --exclude-words to --omit-words, now defaulting to non-speech w/o hesitations.
#   1.5.1 (06 Jan 2022):
#   - Minor fix to mod9-asr-switchboard-benchmark.
#   1.5.0 (06 Jan 2022):
#   - Fixed handling of gs:// URIs in which the blob name requires percent-encoded URLs.
#   - Make "languageCode" optional, in contrast to Google; default is first Engine model loaded.
#   - Do not accept "command_and_search" as model type.
#   - The speech_mod9 module and REST API now extend support for "asrModel".
#     - This will override Google-compatible "languageCode" and "model", if specified.
#   - The speech_mod9 module and REST API now extend support for "maxWordAlternatives".
#   - Only allow "maxAlternatives" up to 1000 transcript-level alternatives.
#   - Improve determination of WAV files by checking header of HTTP(S) URI files.
#   - Confirm existence of audio via URI to avoid waiting for timeout if audio does not exist.
#   - The speech_mod9 module and REST API now extend support for "intervalsJson".
#   - Remove "enablePhraseConfidence" option from speech_mod9 and REST API; always report biases.
#   - The REST API now always reports 19-digit operation names.
#   - Various minor changes to mod9-asr-switchboard-benchmark:
#     - Default of 0 for --max-expansions, fully-expanded alternatives assuming bugfixed SCTK.
#     - Improved parsing of Google STT-formatted results, and optimization of refiltered CTM.
#     - Added --alternatives-max to allow scoring variable depth lists of alternatives.
#     - Added --verbose option, false by default since the tool was otherwise too verbose.
#     - When verbose, report statistics about the size, depth, and width of alternatives.
#   1.4.2 (16 Dec 2021):
#   - Enable mod9-asr-websocket-client to request non-recognize commands without audio data.
#   1.4.1 (27 Nov 2021):
#   - Minor bugfixes.
#   1.4.0 (23 Nov 2021):
#   - Add mod9-asr-switchboard-benchmark to replicate results at remeeting.com/benchmark.
#   1.3.0 (18 Nov 2021):
#   - Add mod9-asr-elasticsearch-client to demonstrate indexing of phrase alternatives.
#   - Enable non-English languages to be specified with "languageCode" option.
#     Unlike Google STT, a region suffix may be omitted, e.g. "en" instead of "en-US".
#   - Support the "model" option, similarly to Google STT.
#   1.2.1 (11 Nov 2021):
#   - Bugfix to allow WebSocket server to handle responses up to 1 MiB (instead of 64KiB).
#   - This setting may be overriden with the --websocket-limit-bytes option.
#   1.2.0 (30 Aug 2021):
#   - Improved logging.
#   - Allow "rate" option to be in the range [8000,48000], as with Google STT.
#   - Added "speed" option to speech_mod9.
#   - Added "options_json" to speech_mod9.
#   1.1.1 (11 Aug 2021):
#   - Rebuild correctly (after `rm -rf build/ dist/ *.egg-info`)
#   1.1.0 (11 Aug 2021):
#   - Released in coordination with Engine version 1.1.0 (coincidental version match, not causal).
#   - Added "latency" request option to speech_mod9.
#   - REST API now logs to a file, with UUIDs both for itself and the proxied Engine.
#   1.0.0 (31 Jul 2021):
#   - This version is not compatible with Engine version < 1.0.0 (due to "asr-model" option).
#   - Bugfixes to WebSocket interface; also add --skip-engine-check and --allow-*-uri (for REST).
#   0.5.0 (28 May 2021): Add Websocket Interface.
#   0.4.1 (20 May 2021): Additional minor documentation fixes; Flask-RESTful version pinning.
#   0.4.0 (30 Apr 2021): Rename mod9-rest-server to mod9-asr-rest-api; minor documentation fixes.

# Range of compatible Engine versions for current wrappers.
#  Lower bound is inclusive, upper bound is exclusive.
#  ``None`` indicates no bound.
# PySDK 1.11.0 requires Engine 1.9.0+ to support multichannel audio.
WRAPPER_ENGINE_COMPATIBILITY_RANGE = ('1.9.0', None)

ASR_ENGINE_HOST = os.getenv('ASR_ENGINE_HOST', 'localhost')
ASR_ENGINE_PORT = int(os.getenv('ASR_ENGINE_PORT', 9900))

SOCKET_CONNECTION_TIMEOUT_SECONDS = 10.0
SOCKET_INACTIVITY_TIMEOUT_SECONDS = 300.0
ENGINE_CONNECTION_RETRY_SECONDS = 1.0

# These should be small enough so that it doesn't trigger the Engine's read timeout (10s default).
MAX_CHUNK_SIZE = 128 * 1024  # Used as chunk size for URI producers; limits generators.
GS_CHUNK_SIZE = 262144  # Google requires chunks be multiples of 262144

FLASK_ENV = os.getenv('FLASK_ENV', None)

# Audio URI prefixes to accept, used by REST only (PySDK allows all).
#  Operator can set at server launch; default is allow none.
ASR_REST_API_ALLOWED_URI_SCHEMES = os.getenv('ASR_REST_API_ALLOWED_URI_SCHEMES', set())
if ASR_REST_API_ALLOWED_URI_SCHEMES:
    ASR_REST_API_ALLOWED_URI_SCHEMES = ASR_REST_API_ALLOWED_URI_SCHEMES.lower().split(sep=',')
    ASR_REST_API_ALLOWED_URI_SCHEMES = set(
        scheme.replace('://', '') for scheme in ASR_REST_API_ALLOWED_URI_SCHEMES
    )

if 'http' in ASR_REST_API_ALLOWED_URI_SCHEMES and 'https' not in ASR_REST_API_ALLOWED_URI_SCHEMES:
    logging.warning('REST API set to allow http:// but NOT https:// audio URIs.')

# Limit on number of bytes allowed per reply line read by WebSocket server.
WEBSOCKET_LIMIT_BYTES = 1024 * 1024  # 1 MiB



#!/usr/bin/env python3

"""
Mod9 ASR REST API: wrapper over the ASR Engine.

This server implements a RESTful API compatible with Google Cloud STT.
It can run as a standalone Flask server, but is best deployed via WSGI.
"""

import argparse
from collections import OrderedDict
from datetime import datetime
import itertools
import logging
import logging.config
from random import randint
import threading
import uuid

from flask import Flask
from flask import Response
from flask_restful import (
    reqparse,
    abort,
    Api,
    Resource,
)

from mod9.asr import speech_mod9
import mod9.reformat.config as config
from mod9.reformat import utils
from mod9.reformat import google as reformat

# Use local time with ISO-8601 format for time output.
logging.Formatter.formatTime = utils.format_log_time

# Configure logging to work well with either WSGI or development webserver.
#  See: https://flask.palletsprojects.com/en/2.0.x/logging/#basic-configuration
logging.config.dictConfig({
    'version': 1,
    'formatters': {
        'default': {
            # For desired formatting, add log level to message below in ``CustomLoggerAdapter``.
            'format': "[%(asctime)s] %(message)s",
        }
    },
    'handlers': {
        'wsgi': {
            'class': 'logging.StreamHandler',
            'stream': 'ext://flask.logging.wsgi_errors_stream',
            'formatter': 'default'
        }
    },
    'root': {
        'level': 'INFO',
        'handlers': ['wsgi']
    }
})


# Use the standard name for a Flask app, at highest scope, to faciliate WSGI integration.
app = Flask(__name__)

# Make JSON output readable.
#  ``RESTFUL_JSON`` is a config var from ``flask_restful`` that
#  allows JSON formatting similar to ``json.dumps()``, see docs:
#  https://flask-restful.readthedocs.io/en/latest/extending.html#:~:text=RESTFUL_JSON
app.config['RESTFUL_JSON'] = {
    'ensure_ascii': False,  # Pass Engine's UTF-8 encoded bytes as-is, not escaped as ASCII.
    'indent': 2,            # Match Google's formatting.
    'sort_keys': False,     # Enable use of OrderedDict, to match Google's formatting.
}

if config.FLASK_ENV is not None:
    app.config['ENV'] = config.FLASK_ENV
else:
    # Stop output of warning message on server start.
    #  Note this is not recommended by Flask devs, but should be fine for our purposes:
    #  https://flask.palletsprojects.com/en/1.1.x/config/#environment-and-debug-features
    app.config['ENV'] = 'development'

# Use Flask-RESTful, a well-designed framework with convenient utilities.
api = Api(app)

# Input parser for Recognize and LongRunningRecognize resources.
parser = reqparse.RequestParser()
parser.add_argument(
    'config',
    required=True,
    help="Required 'config' contains ASR settings in JSON format.",
    location='json',
    type=dict,
)
parser.add_argument(
    'audio',
    required=True,
    help="Required 'audio' contains audio file path or encoded bytes.",
    location='json',
    type=dict,
)

# Storage for operation names and results used by LongRunningRecognize.
operation_names = []
operation_results = {}

# Audio URI schemes to accept. Operator can set at server launch; default is to accept none.
allowed_uri_schemes = config.ASR_REST_API_ALLOWED_URI_SCHEMES


class CustomLoggerAdapter(logging.LoggerAdapter):
    """
    Add ``rest_uuid`` and log level to logs.

    Add log level here so we can format it how we want.

    E.g. see:
    https://docs.python.org/3/howto/logging-cookbook.html#using-loggeradapters-to-impart-contextual-information
    https://github.com/python/cpython/blob/96cf5a63d2dbadaebf236362b4c7c09c51fda55c/Lib/logging/__init__.py#L1799-L1883
    """
    def process(self, msg, kwargs):
        """Prepend [{rest_uuid}:{level}] to message."""
        rest_uuid = self.extra.get('rest_uuid', str(uuid.UUID(int=0)))[:8]
        level = logging.getLevelName(self.extra['level']).lower()
        # Space before tab to avoid jagged logs for, e.g., INFO vs WARNING.
        return f"[{rest_uuid}:{level}] \t{msg}", kwargs

    def log(self, level, msg, *args, **kwargs):
        """Override: pass level as element of ``self.extra`` dict."""
        if self.isEnabledFor(level):
            self.extra['level'] = level
            msg, kwargs = self.process(msg, kwargs)
            self.logger.log(level, msg, *args, **kwargs)


@app.before_first_request
def log_pysdk_version_and_schemes():
    """
    Log server information on start, e.g. version, enabled URI schemes.

    Args:
        None

    Returns:
        None
    """

    logger = CustomLoggerAdapter(logging.root, dict())
    logger.info("Mod9 ASR REST API (Python SDK version %s).", config.WRAPPER_VERSION)

    logger.info(
        "Accepting URI schemes: %s.",
        ', '.join(allowed_uri_schemes) if len(allowed_uri_schemes) else None,
    )

    if 'http' in allowed_uri_schemes and 'https' not in allowed_uri_schemes:
        logger.warning('REST API set to allow http:// but NOT https:// audio URIs.')


def make_request_logger():
    """
    Create a REST UUID and logger for a request.

    Args:
        None

    Returns:
        CustomLoggerAdapter:
            The logger to be used for the request.
    """

    rest_uuid = str(uuid.uuid4())
    logger = CustomLoggerAdapter(logging.root, {'rest_uuid': rest_uuid})
    logger.info("REST UUID: %s.", rest_uuid)
    return logger


def make_response_headers(rest_uuid, engine_uuid=None, engine_version=None):
    """
    Prepare response headers.

    Args:
        rest_uuid (str):
            Request UUID assigned by REST API.
        engine_uuid (Union[str, None]):
            UUID received from Engine for request, or None.
        engine_version (Union[str, None]):
            Version received from Engine as part of response, or None.

    Returns:
        dict:
            The headers.
    """

    headers = {
        'X-Mod9-ASR-REST-API-UUID': rest_uuid,
        'X-Mod9-ASR-REST-API-Version': config.WRAPPER_VERSION,
    }
    if engine_uuid:
        headers['X-Mod9-ASR-Engine-UUID'] = engine_uuid
    if engine_version:
        headers['X-Mod9-ASR-Engine-Version'] = engine_version

    return headers


def place_reformatted_mod9_response_in_operation_results(
    mod9_config_settings,
    mod9_audio_settings,
    operation_name,
    logger,
):
    """
    Get and format response from Mod9 ASR Engine. Place response in
    proper ``operation_result``.

    Args:
        mod9_config_settings (dict[str, Union[dict, float, int, str]]):
            Mod9-style transcription options.
        mod9_audio_settings (Union[
                GeneratorType,
                TeeGeneratorType,
                str,
        ]):
            Audio content to be transcribed.
        operation_name (str):
            Transcription operation UUID.

    Returns:
        None
    """

    # Response type metadata constant string.
    response_type = 'type.googleapis.com/google.cloud.speech.v1p1beta1.LongRunningRecognizeResponse'

    # Talk to Mod9 ASR Engine.
    try:
        engine_response = utils.get_transcripts_mod9(mod9_config_settings, mod9_audio_settings)
    except (KeyError, ConnectionError):
        logger.exception('Error communicating with Mod9 ASR Engine.')
        abort(Response(status=500, headers=make_response_headers(logger.extra['rest_uuid'])))

    # Peek at first Engine response line containing Engine UUID and version.
    first_response_line = next(engine_response)
    engine_response = itertools.chain([first_response_line], engine_response)
    engine_uuid = first_response_line.get('uuid')
    engine_version = first_response_line.get('engine', dict()).get('version')
    logger.info(
        "Engine version %s replied with UUID: %s.",
        engine_version,
        engine_uuid,
    )

    # Place response in operation_results.
    operation_results[operation_name]['response'] = OrderedDict(
        [
            ('@type', response_type),
            ('results', list(reformat.result_from_mod9(engine_response, logger=logger))),
        ]
    )
    operation_results[operation_name]['done'] = True
    operation_results[operation_name].move_to_end('response')

    # Hack to remove .isFinal field from response.
    [result.pop('isFinal') for result in operation_results[operation_name]['response']['results']]

    # Update metadata.
    current_time = datetime.utcnow().isoformat() + 'Z'
    operation_results[operation_name]['metadata']['lastUpdateTime'] = current_time
    operation_results[operation_name]['metadata']['progressPercent'] = 100
    operation_results[operation_name]['metadata'].move_to_end('startTime')
    operation_results[operation_name]['metadata'].move_to_end('lastUpdateTime')

    logger.info('Request completed.')


class Recognize(Resource):
    """Implement synchronous ASR for ``/recognize`` endpoint."""

    def post(self):
        """Perform synchronous ASR on POSTed config and audio."""
        logger = make_request_logger()

        # Translate request -> Mod9 format.
        args = parser.parse_args()
        try:
            mod9_config_settings, mod9_audio_settings = reformat.input_to_mod9(
                args,
                module=speech_mod9,
                logger=logger,
            )
        except Exception as e:
            logger.exception('Invalid arguments.')
            abort(
                Response(
                    status=400,
                    response=f"Invalid arguments: {e}",
                    headers=make_response_headers(logger.extra['rest_uuid']),
                ),
            )

        if isinstance(mod9_audio_settings, str):
            try:
                utils.validate_uri_scheme(mod9_audio_settings, allowed_uri_schemes)
            except utils.Mod9DisabledAudioURISchemeError as e:
                logger.error(e)
                abort(
                    Response(
                        status=403,
                        response=str(e) + '\n',
                        headers=make_response_headers(logger.extra['rest_uuid']),
                    ),
                )

        # Talk to Mod9 ASR Engine.
        try:
            engine_response = utils.get_transcripts_mod9(mod9_config_settings, mod9_audio_settings)
        except (KeyError, ConnectionError):
            logger.exception('Error communicating with Mod9 ASR Engine.')
            abort(Response(status=500, headers=make_response_headers(logger.extra['rest_uuid'])))

        # Peek at first Engine response line containing Engine UUID and version.
        first_response_line = next(engine_response)
        engine_response = itertools.chain([first_response_line], engine_response)
        engine_uuid = first_response_line.get('uuid')
        engine_version = first_response_line.get('engine', dict()).get('version')
        logger.info(
            "Engine version %s replied with UUID: %s.",
            engine_version,
            engine_uuid,
        )

        # Translate response -> external API format.
        response = {
            'results': list(reformat.result_from_mod9(engine_response, logger=logger))
        }

        # Hack to remove .isFinal field from response.
        [result.pop('isFinal') for result in response['results']]

        logger.info('Request completed.')

        headers = make_response_headers(logger.extra['rest_uuid'], engine_uuid, engine_version)
        return response, 200, headers


class Operations(Resource):
    """
    Implement operations name fetching for ``/operations`` endpoint.
    """

    def get(self):
        """Output finished and presently running operation names."""
        logger = make_request_logger()

        logger.info('Fetched operations.')
        headers = make_response_headers(logger.extra['rest_uuid'])
        return {'operations': list(reversed(operation_names))}, 200, headers


class GetOperationByName(Resource):
    """
    Implement operation result fetching for
    ``/operations/<operation_name>`` endpoint.
    """

    def get(self, operation_name):
        """Output operation with given operation name."""
        logger = make_request_logger()

        logger.info("Fetched operation %s.", operation_name)
        headers = make_response_headers(logger.extra['rest_uuid'])
        return operation_results[str(operation_name)], 200, headers


class LongRunningRecognize(Resource):
    """
    Implement asynchronous ASR for ``/longrunningrecognize`` endpoint.
    """

    def post(self):
        """Perform asynchronous ASR on POSTed config and audio."""
        logger = make_request_logger()

        # Translate request -> Mod9 format.
        args = parser.parse_args()
        try:
            mod9_config_settings, mod9_audio_settings = reformat.input_to_mod9(
                args,
                module=speech_mod9,
                logger=logger,
            )
        except Exception as e:
            logger.exception('Invalid arguments.')
            abort(
                Response(
                    status=400,
                    response=f"Invalid arguments: {e}",
                    headers=make_response_headers(logger.extra['rest_uuid']),
                ),
            )

        if isinstance(mod9_audio_settings, str):
            try:
                utils.validate_uri_scheme(mod9_audio_settings, allowed_uri_schemes)
            except utils.Mod9DisabledAudioURISchemeError as e:
                logger.error(e)
                abort(
                    Response(
                        status=403,
                        response=str(e) + '\n',
                        headers=make_response_headers(logger.extra['rest_uuid']),
                    ),
                )

        # Generate a name for operation and append to list of operations.
        # Similar to Google STT, make this a 19-character string of random integers.
        operation_name = str(randint(1, 9)) + ''.join(str(randint(0, 9)) for _ in range(18))
        operation_names.append({'name': operation_name})

        # Set up initial operation_result.
        # Reponse type metadata constant string.
        meta_type = 'type.googleapis.com/google.cloud.speech.v1p1beta1.LongRunningRecognizeMetadata'
        start_time = datetime.utcnow().isoformat() + 'Z'
        operation_metadata = OrderedDict(
            [
                ('@type', meta_type),
                ('startTime', start_time),
                ('lastUpdateTime', start_time),
            ]
        )

        operation_results[operation_name] = OrderedDict([('name', operation_name)])
        operation_results[operation_name]['metadata'] = operation_metadata

        # Send request to Mod9 ASR Engine. Result will appear in operation_results when done.
        request_thread = threading.Thread(
            target=place_reformatted_mod9_response_in_operation_results,
            args=(
                mod9_config_settings,
                mod9_audio_settings,
                operation_name,
                logger,
            ),
        )
        request_thread.start()

        headers = make_response_headers(logger.extra['rest_uuid'])
        return {'name': operation_name}, 200, headers


api.add_resource(
    Recognize,
    '/speech:recognize',
    '/speech:recognize/',
)
api.add_resource(
    LongRunningRecognize,
    '/speech:longrunningrecognize',
    '/speech:longrunningrecognize/',
)
api.add_resource(
    Operations,
    '/operations',
    '/operations/',
)
api.add_resource(
    GetOperationByName,
    '/operations/<int:operation_name>',
    '/operations/<int:operation_name>/',
)


def main():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        '--engine-host',
        help='ASR Engine host name.'
             ' Can also be set by ASR_ENGINE_HOST environment variable.',
        default=config.ASR_ENGINE_HOST,
    )
    parser.add_argument(
        '--engine-port',
        help='ASR Engine port.'
             ' Can also be set by ASR_ENGINE_PORT environment variable.',
        type=int,
        default=config.ASR_ENGINE_PORT,
    )
    parser.add_argument(
        '--host',
        help='REST API host address. Can be set to 0.0.0.0 for external access.',
        default='127.0.0.1'  # Internal access only.
    )
    parser.add_argument(
        '--port',
        help='REST API port number.',
        type=int,
        default=8080,
    )
    parser.add_argument(
        '--log-level',
        metavar='LEVEL',
        help='Verbosity of logging.',
        default='INFO',
    )
    parser.add_argument(
        '--skip-engine-check',
        action='store_true',
        help='When starting server, do not wait for ASR Engine.',
        default=False,
    )
    uri_group = parser.add_argument_group(
        'Audio URI schemes',
        'Allow clients to input audio using a variety of URI schemes.',
    )
    uri_group.add_argument(
        '--allow-file-uri',
        help='Allow clients to use `file://` URI scheme.',
        action='store_const',
        const='file',
    )
    uri_group.add_argument(
        '--allow-gs-uri',
        help='Allow clients to use Google `gs://` URI scheme.',
        action='store_const',
        const='gs',
    )
    uri_group.add_argument(
        '--allow-http-uri',
        help='Allow clients to use `http://` URI scheme.',
        action='store_const',
        const='http',
    )
    uri_group.add_argument(
        '--allow-https-uri',
        help='Allow clients to use `https://` URI scheme.',
        action='store_const',
        const='https',
    )
    uri_group.add_argument(
        '--allow-s3-uri',
        help='Allow clients to use AWS `s3://` URI scheme.',
        action='store_const',
        const='s3',
    )
    args = parser.parse_args()

    global allowed_uri_schemes
    allowed_uri_schemes = allowed_uri_schemes.union(
        {
            args.allow_file_uri,
            args.allow_gs_uri,
            args.allow_http_uri,
            args.allow_https_uri,
            args.allow_s3_uri,
        }
    )
    if None in allowed_uri_schemes:
        allowed_uri_schemes.remove(None)

    logger = CustomLoggerAdapter(logging.root, dict())
    logger.setLevel(args.log_level.upper())

    config.ASR_ENGINE_HOST = args.engine_host
    config.ASR_ENGINE_PORT = args.engine_port

    if not args.skip_engine_check:
        utils.test_host_port(logger=logger)

    # See https://flask.palletsprojects.com/en/2.0.x/deploying/index.html
    logger.warning(
        'This REST API is being run directly as a Python Flask server.'
        ' For production deployment, use WSGI.'
    )
    app.run(host=args.host, port=args.port, debug=False)


if __name__ == '__main__':
    main()



"""
A drop-in replacement for Google Cloud STT Python Client Library.

The speech module provides a strict subset of Google Cloud STT
functionality, and uses Google's provided objects directly. Input in
Google-format objects (or dict) is converted internally to Mod9-format
JSON. Output is converted from Mod9-format JSON to Google-format
objects.
"""

import google.cloud.speech_v1p1beta1 as cloud_speech

from mod9.asr.common import (
    long_running_recognize,
    recognize,
    streaming_recognize,
)
from mod9.reformat import config

__version__ = config.WRAPPER_VERSION

# Copy Google's types into this namespace.
# Not all are used/implemented by Mod9 ASR; they may be silently ignored.
LongRunningRecognizeMetadata = cloud_speech.LongRunningRecognizeMetadata
LongRunningRecognizeRequest = cloud_speech.LongRunningRecognizeRequest
LongRunningRecognizeResponse = cloud_speech.LongRunningRecognizeResponse
RecognitionAudio = cloud_speech.RecognitionAudio
RecognitionConfig = cloud_speech.RecognitionConfig
RecognitionMetadata = cloud_speech.RecognitionMetadata
RecognizeRequest = cloud_speech.RecognizeRequest
RecognizeResponse = cloud_speech.RecognizeResponse
SpeakerDiarizationConfig = cloud_speech.SpeakerDiarizationConfig
SpeechContext = cloud_speech.SpeechContext
SpeechRecognitionAlternative = cloud_speech.SpeechRecognitionAlternative
SpeechRecognitionResult = cloud_speech.SpeechRecognitionResult
StreamingRecognitionConfig = cloud_speech.StreamingRecognitionConfig
StreamingRecognitionResult = cloud_speech.StreamingRecognitionResult
StreamingRecognizeRequest = cloud_speech.StreamingRecognizeRequest
StreamingRecognizeResponse = cloud_speech.StreamingRecognizeResponse
WordInfo = cloud_speech.WordInfo


class Mod9ASREngineTransport(object):
    """Duck-typed SpeechTransport"""
    _wrapped_methods = {}

    def __init__(self, host=None, port=None):
        self._wrapped_methods[self.recognize] = \
            self.recognize
        self._wrapped_methods[self.streaming_recognize] = \
            self.streaming_recognize
        self.host = host
        self.port = port

    def long_running_recognize(self, *args, **kwargs):
        """
        Performs asynchronous speech recognition.

        Mod9: Not currently implemented.
        """
        return long_running_recognize(
            *args,
            host=self.host,
            port=self.port,
            **kwargs,
        )

    def recognize(self, *args, **kwargs):
        """
        Performs synchronous speech recognition: receive results after
        all audio has been sent and processed.

        Args:
            request (Union[dict, RecognizeRequest]):
                The request object. The top-level message sent by the
                client for the `Recognize` method.
            config (RecognitionConfig):
                Required. Provides information to the recognizer that
                specifies how to process the request. This corresponds
                to the ``config`` field on the ``request`` instance; if
                ``request`` is provided, thi should not be set.
            audio (RecognitionAudio):
                Required. The audio data to be recognized. This
                corresponds to the ``audio`` field on the ``request``
                instance; if ``request`` is provided, this should not
                be set.

        Returns:
            RecognizeResponse:
                The only message returned to the client by the
                ``Recognize`` method. It contains the result as zero or
                more sequential ``SpeechRecognitionResult`` messages.
        """
        return recognize(
            *args,
            host=self.host,
            port=self.port,
            **kwargs,
        )

    def streaming_recognize(self, *args, **kwargs):
        """
        Perform streaming speech recognition.

        This method allows you to receive results while sending audio.

        Example:
          >>> from mod9.asr import speech
          >>> client = speech.SpeechClient()
          >>> encoding = speech.RecognitionConfig.AudioEncoding.LINEAR16
          >>> config = speech.StreamingRecognitionConfig(
          ...     config=speech.RecognitionConfig(
          ...         encoding=encoding,
          ...     ),
          ... )
          >>> # Byte-encoded content goes in ``audio_content`` below:
          >>> request = speech.StreamingRecognizeRequest(audio_content=b'')
          >>> requests = [request]
          >>> for element in client.streaming_recognize(config, requests):
          ...     # process element
          ...     pass

        Args:
            config (StreamingRecognitionConfig):
                The configuration to use for the stream.
            requests (Iterable[StreamingRecognizeRequest]):
                The input objects.

        Returns:
            Iterable[StreamingRecognizeResponse]
        """
        return streaming_recognize(
            *args,
            host=self.host,
            port=self.port,
            **kwargs,
        )


class SpeechClient(cloud_speech.SpeechClient):
    """OVERRIDE: drop-in replacement for Google Cloud Speech."""

    def __init__(self, host=None, port=None, *args, **kwargs):
        """OVERRIDE: ignore arguments, set Mod9's custom transport."""
        self._transport = Mod9ASREngineTransport(host=host, port=port)
